"""Fixtures to simplify writing unit tests for AiiDA with ``pytest``."""
# AUTO-GENERATED

# fmt: off

from .configuration import (
    aiida_config,
    aiida_config_factory,
    aiida_config_tmp,
    aiida_profile,
    aiida_profile_clean,
    aiida_profile_clean_class,
    aiida_profile_factory,
    aiida_profile_tmp,
)
from .daemon import daemon_client, started_daemon_client, stopped_daemon_client, submit_and_await
from .entry_points import entry_points
from .globals import aiida_manager
from .orm import (
    aiida_code,
    aiida_code_installed,
    aiida_computer,
    aiida_computer_local,
    aiida_computer_ssh,
    aiida_localhost,
    ssh_key,
)
from .storage import config_psql_dos, config_sqlite_dos, postgres_cluster

__all__ = (
    'aiida_code_installed',
    'aiida_code',
    'aiida_computer_local',
    'aiida_computer_ssh',
    'aiida_computer',
    'aiida_config_factory',
    'aiida_config_tmp',
    'aiida_config',
    'aiida_localhost',
    'aiida_manager',
    'aiida_profile_clean_class',
    'aiida_profile_clean',
    'aiida_profile_factory',
    'aiida_profile_tmp',
    'aiida_profile',
    'config_psql_dos',
    'config_sqlite_dos',
    'daemon_client',
    'entry_points',
    'postgres_cluster',
    'ssh_key',
    'started_daemon_client',
    'stopped_daemon_client',
    'submit_and_await',
)


# fmt: on



"""Fixtures to create and load temporary AiiDA configuration directories and profiles."""

from __future__ import annotations

import contextlib
import os
import pathlib
import secrets
import typing as t

import pytest

if t.TYPE_CHECKING:
    from aiida.manage.configuration.config import Config


@pytest.fixture(scope='session')
def aiida_config_factory():
    """Return a factory to create and load a new temporary AiiDA configuration directory.

    The factory is a context manager that returns a loaded :class:`aiida.manage.configuration.config.Config`. It
    requires a path on the local file system where the configuration directory is to be created as an argument. If
    another configuration directory was already loaded that is automatically restored at the end of the context manager.
    This way, any changes made to the configuration during the context are fully temporary and automatically undone
    after the test.

    Usage::

        def test(aiida_config_factory, tmp_path_factory):
            import secrets
            with aiida_config_factory(tmp_path_factory.mktemp(secrets.token_hex(16))) as config:
                yield config

    The factory has the following signature to allow further configuring the profile that is created and loaded:

    :param dirpath: The path to create the configuration directory in.
    :returns `~aiida.manage.configuration.config.Config`: The loaded temporary config.
    """

    @contextlib.contextmanager
    def factory(dirpath: pathlib.Path):
        from aiida.common.exceptions import MissingConfigurationError
        from aiida.manage.configuration import get_config, reset_config, settings

        try:
            current_config = get_config()
        except MissingConfigurationError:
            current_config = None

        current_path_variable = os.environ.get(settings.DEFAULT_AIIDA_PATH_VARIABLE)

        reset_config()

        dirpath_config = dirpath / settings.DEFAULT_CONFIG_DIR_NAME
        os.environ[settings.DEFAULT_AIIDA_PATH_VARIABLE] = str(dirpath_config)
        settings.set_configuration_directory(dirpath_config)
        config = get_config(create=True)

        try:
            yield config
        finally:
            if current_config:
                reset_config()
                settings.set_configuration_directory(pathlib.Path(current_config.dirpath))
                get_config()

            if current_path_variable is None:
                os.environ.pop(settings.DEFAULT_AIIDA_PATH_VARIABLE, None)
            else:
                os.environ[settings.DEFAULT_AIIDA_PATH_VARIABLE] = current_path_variable

    return factory


@pytest.fixture(scope='session')
def aiida_profile_factory():
    """Return a factory to create and load a new temporary AiiDA profile.

    The factory is a context manager that returns a loaded :class:`aiida.manage.configuration.profile.Profile`. It
    requires a :class:`aiida.manage.configuration.config.Config` instance to which the profile is added. If another
    profile was already loaded, that is automatically restored at the end of the context manager. This way, any changes
    made to the profile during the context are fully temporary and automatically undone after the test. The created
    ``Profile`` instance dynamically has the method ``reset_storage`` added which, when called, deletes all content of
    the storage, recreating the default user. The daemon is also stopped if it was running.

    Usage::

        def test(aiida_config_tmp, aiida_profile_factory):
            with aiida_profile_factory(aiida_config_tmp) as profile:
                yield profile

    The factory has the following signature to allow further configuring the profile that is created and loaded:

    :param storage_backend: The storage plugin to use. Defaults to ``core.sqlite_dos``.
    :param storage_config: The configuration to use for the selected storage plugin.
    :param broker_backend: The broker plugin to use. Defaults to defining no broker.
    :param broker_config: The configuration to use for the selected broker plugin.
    :param name: The name of the profile. Defaults to a random string.
    :param name: The email to use for the default user. Defaults to ``test@localhost``.
    :returns `~aiida.manage.configuration.profile.Profile`: The loaded temporary profile.
    """

    @contextlib.contextmanager
    def factory(
        config: 'Config',
        *,
        storage_backend: str = 'core.sqlite_dos',
        storage_config: dict[str, t.Any] | None = None,
        broker_backend: str | None = None,
        broker_config: dict[str, t.Any] | None = None,
        name: str | None = None,
        email: str = 'test@localhost',
    ):
        from aiida.manage.configuration import create_profile, profile_context
        from aiida.manage.manager import get_manager

        manager = get_manager()
        name = name or secrets.token_hex(16)
        storage_config = storage_config or {'filepath': str(pathlib.Path(config.dirpath) / name / 'storage')}

        if broker_backend and broker_config is None:
            broker_config = {
                'broker_protocol': 'amqp',
                'broker_username': 'guest',
                'broker_password': 'guest',
                'broker_host': '127.0.0.1',
                'broker_port': 5672,
                'broker_virtual_host': '',
            }

        profile = create_profile(
            config,
            storage_backend=storage_backend,
            storage_config=storage_config,
            broker_backend=broker_backend,
            broker_config=broker_config,
            name=name,
            email=email,
            is_test_profile=True,
        )
        config.set_default_profile(profile.name)

        def reset_storage():
            """Reset the storage of the profile.

            This ensures that the contents of the profile are reset as well as the ``Manager``, which may hold
            references to data that will be destroyed. The daemon will also be stopped if it was running.
            """
            from aiida.engine.daemon.client import DaemonException, get_daemon_client
            from aiida.orm import User

            if broker_backend:
                daemon_client = get_daemon_client()

                if daemon_client.is_daemon_running:
                    try:
                        daemon_client.stop_daemon(wait=True)
                    except DaemonException:
                        pass

            manager.get_profile_storage()._clear()
            manager.reset_profile()

            User(email=profile.default_user_email or email).store()

        # Add the ``reset_storage`` method, such that users can empty the storage through the ``Profile`` instance that
        # is returned by this fixture.
        setattr(profile, 'reset_storage', reset_storage)

        with profile_context(profile, allow_switch=True):
            yield profile

    return factory


@pytest.fixture(scope='session', autouse=True)
def aiida_config(tmp_path_factory, aiida_config_factory):
    """Return a loaded temporary AiiDA configuration directory.

    This fixture is session-scoped and used automatically as soon as these fixtures are imported.

    :returns :class:`~aiida.manage.configuration.config.Config`: The loaded temporary config.
    """
    with aiida_config_factory(tmp_path_factory.mktemp(secrets.token_hex(16))) as config:
        yield config


@pytest.fixture(scope='session', autouse=True)
def aiida_profile(aiida_config, aiida_profile_factory):
    """Return a loaded temporary AiiDA profile.

    This fixture is session-scoped and used automatically as soon as these fixtures are imported. The profile defines
    no broker and uses the ``core.sqlite_dos`` storage backend, meaning it requires no services to run.

    :returns :class:`~aiida.manage.configuration.profile.Profile`: The loaded temporary profile.
    """
    with aiida_profile_factory(aiida_config) as profile:
        yield profile


@pytest.fixture(scope='function')
def aiida_profile_clean(aiida_profile):
    """Return a loaded temporary AiiDA profile where the data storage is cleaned before the start of the test.

    This is a function-scoped version of the ``aiida_profile`` fixture.

    :returns :class:`~aiida.manage.configuration.profile.Profile`: The loaded temporary profile.
    """
    aiida_profile.reset_storage()
    yield aiida_profile


@pytest.fixture(scope='class')
def aiida_profile_clean_class(aiida_profile):
    """Return a loaded temporary AiiDA profile where the data storage is cleaned before the start of the test.

    This is a class-scoped version of the ``aiida_profile`` fixture.

    :returns `~aiida.manage.configuration.profile.Profile`: The loaded temporary profile.
    """
    aiida_profile.reset_storage()
    yield aiida_profile


@pytest.fixture(scope='function')
def aiida_config_tmp(tmp_path, aiida_config_factory):
    """Create and load a temporary AiiDA configuration directory.

    This fixture is function-scoped and automatically restores any previously loaded config after the test.

    :returns :class:`~aiida.manage.configuration.config.Config`: The loaded temporary config.
    """
    with aiida_config_factory(tmp_path) as config:
        yield config


@pytest.fixture(scope='function')
def aiida_profile_tmp(aiida_config_tmp, aiida_profile_factory):
    """Create and load a temporary AiiDA profile.

    This fixture is function-scoped and automatically restores any previously loaded profile after the test.

    :returns :class:`~aiida.manage.configuration.profile.Profile`: The loaded temporary profile.
    """
    with aiida_profile_factory(aiida_config_tmp) as profile:
        yield profile



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi process` command."""

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.common.log import LOG_LEVELS, capture_logging

REPAIR_INSTRUCTIONS = """\
If one ore more processes are unreachable, you can run the following commands to try and repair them:

    verdi daemon stop
    verdi process repair
    verdi daemon start
"""


def valid_projections():
    """Return list of valid projections for the ``--project`` option of ``verdi process list``.

    This indirection is necessary to prevent loading the imported module which slows down tab-completion.
    """
    from aiida.tools.query.calculation import CalculationQueryBuilder

    return CalculationQueryBuilder.valid_projections


def default_projections():
    """Return list of default projections for the ``--project`` option of ``verdi process list``.

    This indirection is necessary to prevent loading the imported module which slows down tab-completion.
    """
    from aiida.tools.query.calculation import CalculationQueryBuilder

    return CalculationQueryBuilder.default_projections


def get_most_recent_node():
    """Return the most recent process node.

    :returns: The ``ProcessNode`` with the latest ``ctime``.
    """
    from aiida.orm import ProcessNode, QueryBuilder

    process = QueryBuilder().append(ProcessNode, tag='n').order_by({'n': {'ctime': 'desc'}}).first(flat=True)
    echo.echo_info(f'Most recent node matched: `{process}`.')
    return process


@verdi.group('process')
def verdi_process():
    """Inspect and manage processes."""


@verdi_process.command('list')
@options.PROJECT(type=types.LazyChoice(valid_projections), default=lambda: default_projections())
@options.ORDER_BY()
@options.ORDER_DIRECTION()
@options.GROUP(help='Only include entries that are a member of this group.')
@options.ALL(help='Show all entries, regardless of their process state.')
@options.PROCESS_STATE()
@options.PROCESS_LABEL()
@options.PAUSED()
@options.EXIT_STATUS()
@options.FAILED()
@options.PAST_DAYS()
@options.LIMIT()
@options.RAW()
@click.pass_context
@decorators.with_dbenv()
def process_list(
    ctx,
    all_entries,
    group,
    process_state,
    process_label,
    paused,
    exit_status,
    failed,
    past_days,
    limit,
    project,
    raw,
    order_by,
    order_dir,
):
    """Show a list of processes.

    By default, only processes that are still running are shown, but there are options to show also the finished ones.
    """
    from tabulate import tabulate

    from aiida.cmdline.commands.cmd_daemon import execute_client_command
    from aiida.cmdline.utils.common import print_last_process_state_change
    from aiida.common.exceptions import ConfigurationError
    from aiida.engine.daemon.client import get_daemon_client
    from aiida.orm import ProcessNode, QueryBuilder
    from aiida.tools.query.calculation import CalculationQueryBuilder

    relationships = {}

    if group:
        relationships['with_node'] = group

    builder = CalculationQueryBuilder()
    filters = builder.get_filters(all_entries, process_state, process_label, paused, exit_status, failed)
    query_set = builder.get_query_set(
        relationships=relationships, filters=filters, order_by={order_by: order_dir}, past_days=past_days, limit=limit
    )
    projected = builder.get_projected(query_set, projections=project)
    headers = projected.pop(0)

    if raw:
        tabulated = tabulate(projected, tablefmt='plain')
        echo.echo(tabulated)
        return

    tabulated = tabulate(projected, headers=headers)
    echo.echo(tabulated)
    echo.echo(f'\nTotal results: {len(projected)}\n')

    if 'cached' in project:
        echo.echo_report('\u267b Processes marked with check-mark were not run but taken from the cache.')
        echo.echo_report('Add the option `-P pk cached_from` to the command to display cache source.')

    print_last_process_state_change()

    try:
        client = get_daemon_client()
    except ConfigurationError:
        echo.echo_warning('This profile does not have a broker and so it has no daemon.')
        return

    if not client.is_daemon_running:
        echo.echo_warning('The daemon is not running', bold=True)
        return

    echo.echo_report('Checking daemon load... ', nl=False)
    response = execute_client_command('get_numprocesses')

    if not response:
        # Daemon could not be reached
        return

    try:
        active_workers = response['numprocesses']
    except KeyError:
        echo.echo_report('No active daemon workers.')
    else:
        # Second query to get active process count. Currently this is slow but will be fixed with issue #2770. It is
        # placed at the end of the command so that the user can Ctrl+C after getting the process table.
        slots_per_worker = ctx.obj.config.get_option('daemon.worker_process_slots', scope=ctx.obj.profile.name)
        active_processes = (
            QueryBuilder()
            .append(ProcessNode, filters={'attributes.process_state': {'in': ('created', 'waiting', 'running')}})
            .count()
        )
        available_slots = active_workers * slots_per_worker

        if active_workers == 0:
            echo.echo_warning('The daemon has no active workers!')
            echo.echo_warning('Increase the number of workers with `verdi daemon incr`.')
            return

        percent_load = active_processes / available_slots

        if percent_load > 0.9:  # 90%
            echo.echo_warning(f'{percent_load * 100:.0f}% of the available daemon worker slots have been used!')
            echo.echo_warning('Increase the number of workers with `verdi daemon incr`.')
        else:
            echo.echo_report(f'Using {percent_load * 100:.0f}% of the available daemon worker slots.')


@verdi_process.command('show')
@arguments.PROCESSES()
@options.MOST_RECENT_NODE()
@decorators.with_dbenv()
def process_show(processes, most_recent_node):
    """Show details of processes.

    Show details for one or multiple processes."""
    from aiida.cmdline.utils.common import get_node_info

    if not processes and not most_recent_node:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and most_recent_node:
        raise click.BadOptionUsage(
            'most_recent_node',
            'cannot specify individual processes and the `-M/--most-recent-node` flag at the same time.',
        )

    if most_recent_node:
        processes = [get_most_recent_node()]

    for process in processes:
        echo.echo(get_node_info(process))


@verdi_process.command('call-root')
@arguments.PROCESSES()
@decorators.with_dbenv()
def process_call_root(processes):
    """Show root process of processes.

    Show root process(es) of the call stack for one or multiple processes."""
    if not processes:
        raise click.UsageError('Please specify one or multiple processes by their identifier (PK, UUID or label).')
    for process in processes:
        caller = process.caller

        if caller is None:
            echo.echo(f'No callers found for Process<{process.pk}>')
            continue

        while True:
            next_caller = caller.caller

            if next_caller is None:
                break

            caller = next_caller

        echo.echo(f'{caller.pk}')


@verdi_process.command('report')
@arguments.PROCESSES()
@options.MOST_RECENT_NODE()
@click.option('-i', '--indent-size', type=int, default=2, help='Set the number of spaces to indent each level by.')
@click.option(
    '-l',
    '--levelname',
    type=click.Choice(list(LOG_LEVELS)),
    default='REPORT',
    help='Filter the results by name of the log level.',
)
@click.option(
    '-m', '--max-depth', 'max_depth', type=int, default=None, help='Limit the number of levels to be printed.'
)
@decorators.with_dbenv()
def process_report(processes, most_recent_node, levelname, indent_size, max_depth):
    """Show the log report of processes.

    Show the log report for one or multiple processes."""
    from aiida.cmdline.utils.common import get_calcjob_report, get_process_function_report, get_workchain_report
    from aiida.orm import CalcFunctionNode, CalcJobNode, WorkChainNode, WorkFunctionNode

    if not processes and not most_recent_node:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and most_recent_node:
        raise click.BadOptionUsage(
            'most_recent_node',
            'cannot specify individual processes and the `-M/--most-recent-node` flag at the same time.',
        )

    if most_recent_node:
        processes = [get_most_recent_node()]

    for process in processes:
        if isinstance(process, CalcJobNode):
            echo.echo(get_calcjob_report(process))
        elif isinstance(process, WorkChainNode):
            echo.echo(get_workchain_report(process, levelname, indent_size, max_depth))
        elif isinstance(process, (CalcFunctionNode, WorkFunctionNode)):
            echo.echo(get_process_function_report(process))
        else:
            echo.echo(f'Nothing to show for node type {process.__class__}')


@verdi_process.command('status')
@options.MOST_RECENT_NODE()
@click.option('-c', '--call-link-label', 'call_link_label', is_flag=True, help='Include the call link label if set.')
@click.option(
    '-m', '--max-depth', 'max_depth', type=int, default=None, help='Limit the number of levels to be printed.'
)
@arguments.PROCESSES()
def process_status(call_link_label, most_recent_node, max_depth, processes):
    """Show the status of processes.

    Show the status of one or multiple processes."""
    from aiida.cmdline.utils.ascii_vis import format_call_graph

    if not processes and not most_recent_node:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and most_recent_node:
        raise click.BadOptionUsage(
            'most_recent_node',
            'cannot specify individual processes and the `-M/--most-recent-node` flag at the same time.',
        )

    if most_recent_node:
        processes = [get_most_recent_node()]

    for process in processes:
        graph = format_call_graph(process, max_depth=max_depth, call_link_label=call_link_label)
        echo.echo(graph)


@verdi_process.command('kill')
@arguments.PROCESSES()
@options.ALL(help='Kill all processes if no specific processes are specified.')
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_kill(processes, all_entries, timeout, wait):
    """Kill running processes.

    Kill one or multiple running processes."""
    from aiida.engine.processes import control

    if not processes and not all_entries:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and all_entries:
        raise click.BadOptionUsage('all', 'cannot specify individual processes and the `--all` flag at the same time.')

    if all_entries:
        click.confirm('Are you sure you want to kill all processes?', abort=True)

    with capture_logging() as stream:
        try:
            message = 'Killed through `verdi process kill`'
            control.kill_processes(processes, all_entries=all_entries, timeout=timeout, wait=wait, message=message)
        except control.ProcessTimeoutException as exception:
            echo.echo_critical(f'{exception}\n{REPAIR_INSTRUCTIONS}')

        if 'unreachable' in stream.getvalue():
            echo.echo_report(REPAIR_INSTRUCTIONS)


@verdi_process.command('pause')
@arguments.PROCESSES()
@options.ALL(help='Pause all active processes if no specific processes are specified.')
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_pause(processes, all_entries, timeout, wait):
    """Pause running processes.

    Pause one or multiple running processes."""
    from aiida.engine.processes import control

    if not processes and not all_entries:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and all_entries:
        raise click.BadOptionUsage('all', 'cannot specify individual processes and the `--all` flag at the same time.')

    with capture_logging() as stream:
        try:
            message = 'Paused through `verdi process pause`'
            control.pause_processes(processes, all_entries=all_entries, timeout=timeout, wait=wait, message=message)
        except control.ProcessTimeoutException as exception:
            echo.echo_critical(f'{exception}\n{REPAIR_INSTRUCTIONS}')

        if 'unreachable' in stream.getvalue():
            echo.echo_report(REPAIR_INSTRUCTIONS)


@verdi_process.command('play')
@arguments.PROCESSES()
@options.ALL(help='Play all paused processes if no specific processes are specified.')
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_play(processes, all_entries, timeout, wait):
    """Play (unpause) paused processes.

    Play (unpause) one or multiple paused processes."""
    from aiida.engine.processes import control

    if not processes and not all_entries:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and all_entries:
        raise click.BadOptionUsage('all', 'cannot specify individual processes and the `--all` flag at the same time.')

    with capture_logging() as stream:
        try:
            control.play_processes(processes, all_entries=all_entries, timeout=timeout, wait=wait)
        except control.ProcessTimeoutException as exception:
            echo.echo_critical(f'{exception}\n{REPAIR_INSTRUCTIONS}')

        if 'unreachable' in stream.getvalue():
            echo.echo_report(REPAIR_INSTRUCTIONS)


@verdi_process.command('watch')
@arguments.PROCESSES()
@options.MOST_RECENT_NODE()
@decorators.with_dbenv()
@decorators.with_broker
@decorators.only_if_daemon_running(echo.echo_warning, 'daemon is not running, so process may not be reachable')
def process_watch(broker, processes, most_recent_node):
    """Watch the state transitions of processes.

    Watch the state transitions for one or multiple running processes."""

    if not processes and not most_recent_node:
        raise click.UsageError(
            'Please specify one or multiple processes by their identifier (PK, UUID or label) or use an option.'
        )

    if processes and most_recent_node:
        raise click.BadOptionUsage(
            'most_recent_node',
            'cannot specify individual processes and the `-M/--most-recent-node` flag at the same time.',
        )

    from time import sleep

    from kiwipy import BroadcastFilter

    def _print(communicator, body, sender, subject, correlation_id):
        """Format the incoming broadcast data into a message and echo it to stdout."""
        if body is None:
            body = 'No message specified'

        if correlation_id is None:
            correlation_id = '--'

        echo.echo(f'Process<{sender}> [{subject}|{correlation_id}]: {body}')

    communicator = broker.get_communicator()
    echo.echo_report('watching for broadcasted messages, press CTRL+C to stop...')

    if most_recent_node:
        processes = [get_most_recent_node()]

    for process in processes:
        if process.is_terminated:
            echo.echo_error(f'Process<{process.pk}> is already terminated')
            continue

        communicator.add_broadcast_subscriber(BroadcastFilter(_print, sender=process.pk))

    try:
        # Block this thread indefinitely until interrupt
        while True:
            sleep(2)
    except (SystemExit, KeyboardInterrupt):
        echo.echo('')  # add a new line after the interrupt character
        echo.echo_report('received interrupt, exiting...')
        try:
            communicator.close()
        except RuntimeError:
            pass

        # Reraise to trigger clicks builtin abort sequence
        raise


@verdi_process.command('repair')
@options.DRY_RUN()
@decorators.only_if_daemon_not_running()
@decorators.with_manager
@decorators.with_broker
def process_repair(manager, broker, dry_run):
    """Automatically repair all stuck processes.

    N.B.: This command requires the daemon to be stopped.

    This command queries the database to find all "active" processes, meaning those that haven't yet reached a terminal
    state, and cross-references them with the active process tasks in the process queue of RabbitMQ. Any active process
    that does not have a corresponding process task can be considered a zombie, as it will never be picked up by a
    daemon worker to complete it and will effectively be "stuck". Any process task that does not correspond to an active
    process is useless and should be discarded. Finally, duplicate process tasks are also problematic and are discarded.
    """
    from aiida.engine.processes.control import get_active_processes, get_process_tasks, iterate_process_tasks

    active_processes = get_active_processes(project='id')
    process_tasks = get_process_tasks(broker)

    set_active_processes = set(active_processes)
    set_process_tasks = set(process_tasks)

    echo.echo_info(f'Active processes: {active_processes}')
    echo.echo_info(f'Process tasks: {process_tasks}')

    state_inconsistent = False

    if len(process_tasks) != len(set_process_tasks):
        state_inconsistent = True
        echo.echo_warning('There are duplicates process tasks: ', nl=False)
        echo.echo(set(x for x in process_tasks if process_tasks.count(x) > 1))

    if set_process_tasks.difference(set_active_processes):
        state_inconsistent = True
        echo.echo_warning('There are process tasks for terminated processes: ', nl=False)
        echo.echo(set_process_tasks.difference(set_active_processes))

    if set_active_processes.difference(set_process_tasks):
        state_inconsistent = True
        echo.echo_warning('There are active processes without process task: ', nl=False)
        echo.echo(set_active_processes.difference(set_process_tasks))

    if not state_inconsistent:
        echo.echo_success('No inconsistencies detected between database and RabbitMQ.')
        return

    echo.echo_warning('Inconsistencies detected between database and RabbitMQ.')

    if dry_run:
        echo.echo_critical('This was a dry-run, no changes will be made.')

    # At this point we have either exited because of inconsistencies and ``--dry-run`` was passed, or we returned
    # because there were no inconsistencies, so all that is left is to address inconsistencies
    echo.echo_report('Attempting to fix inconsistencies')

    # Eliminate duplicate tasks and tasks that correspond to terminated process
    for task in iterate_process_tasks(broker):
        pid = task.body.get('args', {}).get('pid', None)
        if pid not in set_active_processes:
            with task.processing() as outcome:
                outcome.set_result(False)
            echo.echo_report(f'Acknowledged task `{pid}`')

    # Revive zombie processes that no longer have a process task
    process_controller = manager.get_process_controller()
    for pid in set_active_processes:
        if pid not in set_process_tasks:
            process_controller.continue_process(pid)
            echo.echo_report(f'Revived process `{pid}`')


@verdi_process.command('dump')
@arguments.PROCESS()
@options.PATH()
@options.OVERWRITE()
@click.option(
    '--include-inputs/--exclude-inputs',
    default=True,
    show_default=True,
    help='Include the linked input nodes of the `CalculationNode`(s).',
)
@click.option(
    '--include-outputs/--exclude-outputs',
    default=False,
    show_default=True,
    help='Include the linked output nodes of the `CalculationNode`(s).',
)
@click.option(
    '--include-attributes/--exclude-attributes',
    default=True,
    show_default=True,
    help='Include attributes in the `.aiida_node_metadata.yaml` written for every `ProcessNode`.',
)
@click.option(
    '--include-extras/--exclude-extras',
    default=True,
    show_default=True,
    help='Include extras in the `.aiida_node_metadata.yaml` written for every `ProcessNode`.',
)
@click.option(
    '-f',
    '--flat',
    is_flag=True,
    default=False,
    help='Dump files in a flat directory for every step of the workflow.',
)
def process_dump(
    process,
    path,
    overwrite,
    include_inputs,
    include_outputs,
    include_attributes,
    include_extras,
    flat,
) -> None:
    """Dump process input and output files to disk.

    Child calculations/workflows (also called `CalcJob`s/`CalcFunction`s and `WorkChain`s/`WorkFunction`s in AiiDA
    jargon) run by the parent workflow are contained in the directory tree as sub-folders and are sorted by their
    creation time.  The directory tree thus mirrors the logical execution of the workflow, which can also be queried by
    running `verdi process status <pk>` on the command line.

    By default, input and output files of each calculation can be found in the corresponding "inputs" and
    "outputs" directories (the former also contains the hidden ".aiida" folder with machine-readable job execution
    settings). Additional input and output files (depending on the type of calculation) are placed in the "node_inputs"
    and "node_outputs", respectively.

    Lastly, every folder also contains a hidden, human-readable `.aiida_node_metadata.yaml` file with the relevant AiiDA
    node data for further inspection.
    """

    from aiida.tools.dumping.processes import ProcessDumper

    process_dumper = ProcessDumper(
        include_inputs=include_inputs,
        include_outputs=include_outputs,
        include_attributes=include_attributes,
        include_extras=include_extras,
        overwrite=overwrite,
        flat=flat,
    )

    try:
        dump_path = process_dumper.dump(process_node=process, output_path=path)
    except FileExistsError:
        echo.echo_critical(
            'Dumping directory exists and overwrite is False. Set overwrite to True, or delete directory manually.'
        )
    except Exception as e:
        echo.echo_critical(f'Unexpected error while dumping {process.__class__.__name__} <{process.pk}>:\n ({e!s}).')

    echo.echo_success(f'Raw files for {process.__class__.__name__} <{process.pk}> dumped into folder `{dump_path}`.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of caching mechanism and configuration for calculations."""

from __future__ import annotations

import keyword
import re
from collections import namedtuple
from contextlib import contextmanager, suppress
from enum import Enum

from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.manage.configuration import get_config_option
from aiida.plugins.entry_point import ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP, ENTRY_POINT_STRING_SEPARATOR

__all__ = ('get_use_cache', 'enable_caching', 'disable_caching')


class ConfigKeys(Enum):
    """Valid keys for caching configuration."""

    DEFAULT = 'caching.default_enabled'
    ENABLED = 'caching.enabled_for'
    DISABLED = 'caching.disabled_for'


class _ContextCache:
    """Cache options, accounting for when in enable_caching or disable_caching contexts."""

    def __init__(self):
        self._default_all = None
        self._enable = []
        self._disable = []

    def clear(self):
        """Clear caching overrides."""
        self.__init__()  # type: ignore[misc]

    def enable_all(self):
        self._default_all = 'enable'

    def disable_all(self):
        self._default_all = 'disable'

    def enable(self, identifier: str):
        self._enable.append(identifier)
        with suppress(ValueError):
            self._disable.remove(identifier)

    def disable(self, identifier: str):
        self._disable.append(identifier)
        with suppress(ValueError):
            self._enable.remove(identifier)

    def get_options(self, strict: bool = False):
        """Return the options, applying any context overrides.

        :param strict: When set to ``True``, the function will actually try to resolve the identifier by loading it and
            if it fails, an exception is raised.
        """
        if self._default_all == 'disable':
            return False, [], []

        if self._default_all == 'enable':
            return True, [], []

        default = get_config_option(ConfigKeys.DEFAULT.value)
        enabled = get_config_option(ConfigKeys.ENABLED.value)[:]
        disabled = get_config_option(ConfigKeys.DISABLED.value)[:]

        for ident in self._disable:
            disabled.append(ident)
            with suppress(ValueError):
                enabled.remove(ident)

        for ident in self._enable:
            enabled.append(ident)
            with suppress(ValueError):
                disabled.remove(ident)

        # Check validity of enabled and disabled entries
        try:
            for identifier in enabled + disabled:
                _validate_identifier_pattern(identifier=identifier, strict=strict)
        except ValueError as exc:
            raise exceptions.ConfigurationError('Invalid identifier pattern in enable or disable list.') from exc

        return default, enabled, disabled


_CONTEXT_CACHE = _ContextCache()


@contextmanager
def enable_caching(*, identifier: str | None = None, strict: bool = False):
    """Context manager to enable caching, either for a specific node class, or globally.

    .. warning:: this does not affect the behavior of the daemon, only the local Python interpreter.

    :param identifier: Process type string of the node, or a pattern with '*' wildcard that matches it.
        If not provided, caching is enabled for all classes.
    :param strict: When set to ``True``, the function will actually try to resolve the identifier by loading it and if
        it fails, an exception is raised.
    :type identifier: str
    """
    type_check(identifier, str, allow_none=True)

    if identifier is None:
        _CONTEXT_CACHE.enable_all()
    else:
        _validate_identifier_pattern(identifier=identifier, strict=strict)
        _CONTEXT_CACHE.enable(identifier)
    yield
    _CONTEXT_CACHE.clear()


@contextmanager
def disable_caching(*, identifier: str | None = None, strict: bool = False):
    """Context manager to disable caching, either for a specific node class, or globally.

    .. warning:: this does not affect the behavior of the daemon, only the local Python interpreter.

    :param identifier: Process type string of the node, or a pattern with '*' wildcard that matches it.
        If not provided, caching is disabled for all classes.
    :param strict: When set to ``True``, the function will actually try to resolve the identifier by loading it and if
        it fails, an exception is raised.
    :type identifier: str
    """
    type_check(identifier, str, allow_none=True)

    if identifier is None:
        _CONTEXT_CACHE.disable_all()
    else:
        _validate_identifier_pattern(identifier=identifier, strict=strict)
        _CONTEXT_CACHE.disable(identifier)
    yield
    _CONTEXT_CACHE.clear()


def get_use_cache(*, identifier: str | None = None, strict: bool = False) -> bool:
    """Return whether the caching mechanism should be used for the given process type according to the configuration.

    :param identifier: Process type string of the node
    :param strict: When set to ``True``, the function will actually try to resolve the identifier by loading it and if
        it fails, an exception is raised.
    :return: boolean, True if caching is enabled, False otherwise
    :raises: `~aiida.common.exceptions.ConfigurationError` if the configuration is invalid, either due to a general
        configuration error, or by defining the class both enabled and disabled
    """
    type_check(identifier, str, allow_none=True)

    default, enabled, disabled = _CONTEXT_CACHE.get_options(strict=strict)

    if identifier is not None:
        type_check(identifier, str)

        enable_matches = [pattern for pattern in enabled if _match_wildcard(string=identifier, pattern=pattern)]
        disable_matches = [pattern for pattern in disabled if _match_wildcard(string=identifier, pattern=pattern)]

        if enable_matches and disable_matches:
            # If both enable and disable have matching identifier, we search for
            # the most specific one. This is determined by checking whether
            # all other patterns match the specific pattern.
            PatternWithResult = namedtuple('PatternWithResult', ['pattern', 'use_cache'])
            most_specific = []
            for specific_pattern in enable_matches:
                if all(
                    _match_wildcard(string=specific_pattern, pattern=other_pattern)
                    for other_pattern in enable_matches + disable_matches
                ):
                    most_specific.append(PatternWithResult(pattern=specific_pattern, use_cache=True))
            for specific_pattern in disable_matches:
                if all(
                    _match_wildcard(string=specific_pattern, pattern=other_pattern)
                    for other_pattern in enable_matches + disable_matches
                ):
                    most_specific.append(PatternWithResult(pattern=specific_pattern, use_cache=False))

            if len(most_specific) > 1:
                raise exceptions.ConfigurationError(
                    f'Invalid configuration: multiple matches for identifier `{identifier}`, but the most specific '
                    f'identifier is not unique. Candidates: {[match.pattern for match in most_specific]}'
                )
            if not most_specific:
                raise exceptions.ConfigurationError(
                    f'Invalid configuration: multiple matches for identifier `{identifier}`, but none of them is most '
                    'specific.'
                )
            return most_specific[0].use_cache
        if enable_matches:
            return True
        if disable_matches:
            return False
    return default


def _match_wildcard(*, string: str, pattern: str) -> bool:
    """Return whether a given name matches a pattern which can contain '*' wildcards.

    :param string: The string to check.
    :param pattern: The patter to match for.
    :returns: ``True`` if ``string`` matches the ``pattern``, ``False`` otherwise.
    """
    regexp = '.*'.join(re.escape(part) for part in pattern.split('*'))
    return re.fullmatch(pattern=regexp, string=string) is not None


def _validate_identifier_pattern(*, identifier: str, strict: bool = False):
    """Validate an caching identifier pattern.

    The identifier (without wildcards) can have one of two forms:

    1.  <group_name><ENTRY_POINT_STRING_SEPARATOR><tail>

        where `group_name` is one of the keys in `ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP`
        and `tail` can be anything _except_ `ENTRY_POINT_STRING_SEPARATOR`.

    2. a fully qualified Python name

        this is a colon-separated string, where each part satisfies
        `part.isidentifier() and not keyword.iskeyword(part)`

    This function checks if an identifier _with_ wildcards can possibly match one of these two forms. If it can not,
    a ``ValueError`` is raised.

    :param identifier: Process type string, or a pattern with '*' wildcard that matches it.
    :param strict: When set to ``True``, the function will actually try to resolve the identifier by loading it and if
        it fails, an exception is raised.
    :raises ValueError: If the identifier is an invalid identifier.
    :raises ValueError: If ``strict=True`` and the identifier cannot be successfully loaded.
    """
    import importlib

    from aiida.common.exceptions import EntryPointError
    from aiida.plugins.entry_point import load_entry_point_from_string

    common_error_msg = f'Invalid identifier pattern `{identifier}`: '
    assert ENTRY_POINT_STRING_SEPARATOR not in '.*'  # The logic of this function depends on this
    # Check if it can be an entry point string
    if identifier.count(ENTRY_POINT_STRING_SEPARATOR) > 1:
        raise ValueError(
            f'{common_error_msg}Can contain at most one entry point string separator `{ENTRY_POINT_STRING_SEPARATOR}`'
        )
    # If there is one separator, it must be an entry point string.
    # Check if the left hand side is a matching pattern
    if ENTRY_POINT_STRING_SEPARATOR in identifier:
        group_pattern, _ = identifier.split(ENTRY_POINT_STRING_SEPARATOR)
        if not any(
            _match_wildcard(string=group_name, pattern=group_pattern)
            for group_name in ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP
        ):
            raise ValueError(
                common_error_msg
                + f'Group name pattern `{group_pattern}` does not match any of the AiiDA entry point group names.'
            )

        # If strict mode is enabled and the identifier is explicit, i.e., doesn't contain a wildcard, try to load it.
        if strict and '*' not in identifier:
            try:
                load_entry_point_from_string(identifier)
            except EntryPointError as exception:
                raise ValueError(common_error_msg + f'`{identifier}` cannot be loaded.') from exception

        # The group name pattern matches, and there are no further entry point string separators in the identifier,
        # hence it is a valid pattern.
        return

    # The separator might be swallowed in a wildcard, for example
    # aiida.* or aiida.calculations*
    if '*' in identifier:
        group_part, _ = identifier.split('*', 1)
        if any(group_name.startswith(group_part) for group_name in ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP):
            return

    # Finally, check if it could be a fully qualified Python name
    for identifier_part in identifier.split('.'):
        # If it contains a wildcard, we can not check for keywords.
        # Replacing all wildcards with a single letter must give an
        # identifier - this checks for invalid characters, and that it
        # does not start with a number.
        if '*' in identifier_part:
            if not identifier_part.replace('*', 'a').isidentifier():
                raise ValueError(
                    common_error_msg
                    + f'Identifier part `{identifier_part}` can not match a fully qualified Python name.'
                )
        else:
            if not identifier_part.isidentifier():
                raise ValueError(f'{common_error_msg}`{identifier_part}` is not a valid Python identifier.')
            if keyword.iskeyword(identifier_part):
                raise ValueError(f'{common_error_msg}`{identifier_part}` is a reserved Python keyword.')

    if not strict:
        return

    # If there is no separator, it must be a fully qualified Python name.
    try:
        module_name = '.'.join(identifier.split('.')[:-1])
        class_name = identifier.split('.')[-1]
        module = importlib.import_module(module_name)
        getattr(module, class_name)
    except (ModuleNotFoundError, AttributeError, IndexError) as exc:
        raise ValueError(common_error_msg + f'`{identifier}` cannot be imported.') from exc



"""Utilities to define monitor functions for ``CalcJobs``."""

from __future__ import annotations

import collections
import dataclasses
import enum
import inspect
import typing as t
from datetime import datetime, timedelta

from aiida.common.lang import type_check
from aiida.common.log import AIIDA_LOGGER
from aiida.orm import CalcJobNode, Dict
from aiida.plugins import BaseFactory

if t.TYPE_CHECKING:
    from aiida.transports import Transport

LOGGER = AIIDA_LOGGER.getChild(__name__)


class CalcJobMonitorAction(enum.Enum):
    """The action a engine should undertake as a result of a monitor."""

    KILL = 'kill'  # The job should be killed and no other monitors should be called.
    DISABLE_ALL = 'disable-all'  # All monitors should be disabled for the continued duration of the current job
    DISABLE_SELF = 'disable-self'  # Disable the monitor that returns a result with this action.


@dataclasses.dataclass
class CalcJobMonitorResult:
    """Data class representing the result of a monitor."""

    key: str | None = None
    """Key of the monitor in the ``monitors`` input namespace. Will be set automatically by the engine."""

    message: str | None = None
    """Human readable message: could be a warning or an error message."""

    action: CalcJobMonitorAction = CalcJobMonitorAction.KILL
    """The action the engine should take."""

    retrieve: bool = True
    """If set to ``False``, the engine will skip retrieving the output files."""

    parse: bool = True
    """If set to ``False``, the engine will skip the parsing of the retrieved files, if one was specified in inputs."""

    override_exit_code: bool = True
    """If set to ``False``, the engine will keep the exit code returned by the parser."""

    outputs: dict[str, t.Any] | None = None
    """Optional dictionary of output nodes to be attached to the process."""

    def __post_init__(self):
        """Validate the attributes."""
        self.validate()

    def validate(self):
        """Validate the instance.

        :raises TypeError: If any of the attributes are of the incorrect type.
        :raises ValueError: If ``parse == True`` and ``retrieve == False``.
        """
        type_check(self.key, str, allow_none=True)
        type_check(self.message, str, allow_none=True)
        type_check(self.action, CalcJobMonitorAction)
        type_check(self.retrieve, bool)
        type_check(self.parse, bool)
        type_check(self.override_exit_code, bool)

        if self.retrieve is False and self.parse is True:
            raise ValueError('`parse` cannot be `True` if `retrieve` is `False`.')


@dataclasses.dataclass
class CalcJobMonitor:
    """Data class representing a monitor for a ``CalcJob``."""

    entry_point: str
    """Entry point in the ``aiida.calculations.monitors`` group implementing the monitor interface."""

    kwargs: dict[t.Any, t.Any] = dataclasses.field(default_factory=dict)
    """Keyword arguments that will be passed to the monitor when invoked (should be JSON serializable)."""

    priority: int = 0
    """Determines the order in which monitors should be executed in the case of multiple monitors."""

    minimum_poll_interval: int | None = None
    """Optional minimum poll interval. If set, subsequent calls should be at least this many seconds apart."""

    call_timestamp: datetime | None = None
    """Optional datetime representing the last time this monitor was called."""

    disabled: bool = False
    """If this attribute is set to ``True`` the monitor should not be called when monitors are processed."""

    def __post_init__(self):
        """Validate the attributes."""
        self.validate()

    def validate(self):
        """Validate the monitor.

        :raises EntryPointError: If the entry point does not exist or cannot be loaded.
        :raises TypeError: If any of the attributes are of the incorrect type.
        :raises ValueError: If the entry point has an incorrect function signature.
        :raises ValueError: If the kwargs specified are not recognized by the function associated with the entry point.
        """
        type_check(self.entry_point, str)
        type_check(self.kwargs, dict)
        type_check(self.priority, int)
        type_check(self.minimum_poll_interval, int, allow_none=True)
        type_check(self.disabled, bool)

        if self.minimum_poll_interval is not None and self.minimum_poll_interval <= 0:
            raise ValueError('The `minimum_poll_interval` must be a positive integer greater than zero.')

        monitor = self.load_entry_point()
        signature = inspect.signature(monitor)
        parameters = list(signature.parameters.keys())

        if any(required_parameter not in parameters for required_parameter in ('node', 'transport')):
            correct_signature = '(node: CalcJobNode, transport: Transport, **kwargs) str | None:'
            raise ValueError(
                f'The monitor `{self.entry_point}` has an invalid function signature, it should be: {correct_signature}'
            )

        unsupported_kwargs = [kwarg for kwarg in self.kwargs if kwarg not in parameters]

        if unsupported_kwargs:
            raise ValueError(f'The monitor `{self.entry_point}` does not accept the keywords: {unsupported_kwargs}.')

    def load_entry_point(self) -> type:
        """Return the function associated with the entry point of this monitor.

        :raises EntryPointError: If the entry point does not exist or cannot be loaded.
        """
        return BaseFactory('aiida.calculations.monitors', self.entry_point)


class CalcJobMonitors:
    """Collection of ``CalcJobMonitor`` instances.

    The collection is initialized from a dictionary where the values are the parameters for initializing an instance of
    :class:`~aiida.engine.processes.calcjobs.monitors.CalcJobMonitor`, which are stored as an ordered dictionary. The
    monitors are sorted according to the priority set for the monitors (reversed, i.e., from high to low) and second
    alphabetically on their key.

    The :meth:`~aiida.engine.processes.calcjobs.monitors.CalcJobMonitors.process` method can be called providing an
    instance of a ``CalcJobNode`` and a ``Transport`` and it will iterate over the collection of monitors, executing
    each monitor in order, and stopping on the first to return a ``CalcJobMonitorResult`` to pass it up to its caller.
    """

    def __init__(self, monitors: dict[str, Dict]):
        """Construct a new instance."""
        type_check(monitors, dict)

        if any(not isinstance(monitor, Dict) for monitor in monitors.values()):
            raise TypeError('at least one value of `monitors` is not a `Dict` node.')

        monitors = {key: CalcJobMonitor(**node.get_dict()) for key, node in monitors.items()}
        self._monitors = collections.OrderedDict(sorted(monitors.items(), key=lambda x: (-x[1].priority, x[0])))

    @property
    def monitors(self) -> collections.OrderedDict:
        """Return an ordered dictionary of the monitor collection.

        Monitors are first sorted on their priority (reversed, i.e., from high to low) and second on their key.

        :returns: Ordered dictionary of monitors..
        """
        return self._monitors

    def process(
        self,
        node: CalcJobNode,
        transport: Transport,
    ) -> CalcJobMonitorResult | None:
        """Call all monitors in order and return the result as one returns anything other than ``None``.

        :param node: The node to pass to the monitor invocation.
        :param transport: The transport to pass to the monitor invocation.
        :returns: ``None`` or a monitor result.
        """
        for key, monitor in self.monitors.items():
            if monitor.disabled:
                LOGGER.debug(f'monitor`{key}` is disabled, skipping')
                continue

            if (
                monitor.minimum_poll_interval
                and monitor.call_timestamp
                and datetime.now() - monitor.call_timestamp < timedelta(seconds=monitor.minimum_poll_interval)
            ):
                LOGGER.debug(f'skipping monitor `{key}` because minimum poll interval has not expired yet.')
                continue

            monitor_function = monitor.load_entry_point()

            LOGGER.debug(f'calling monitor `{key}`')
            monitor.call_timestamp = datetime.now()
            monitor_result = monitor_function(node, transport, **monitor.kwargs)

            if isinstance(monitor_result, str):
                monitor_result = CalcJobMonitorResult(message=monitor_result)

            if isinstance(monitor_result, CalcJobMonitorResult):
                monitor_result.key = key

            if monitor_result:
                LOGGER.info(f'Monitor `{key}` returned: {monitor_result}')
                return monitor_result



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Rules for the AiiDA Graph Explorer utility"""

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from copy import deepcopy

import numpy as np

from aiida import orm
from aiida.common.lang import type_check
from aiida.tools.graph.age_entities import Basket


class Operation(metaclass=ABCMeta):
    """Base class for all AGE explorer classes"""

    def __init__(self, max_iterations, track_edges):
        """Initialization method

        :param max_iterations: maximum number of iterations to perform.
        :param bool track_edges: if True, will also track and return the edges traversed.
        """
        self.set_max_iterations(max_iterations)
        self._track_edges = track_edges
        self._iterations_done = None

    def set_max_iterations(self, max_iterations):
        """Sets the max iterations"""
        if not (isinstance(max_iterations, int) or max_iterations is np.inf):
            raise TypeError('max_iterations has to be an integer or np.inf')
        self._max_iterations = max_iterations

    @property
    def iterations_done(self):
        """Number of iterations performed"""
        return self._iterations_done

    @abstractmethod
    def run(self, operational_set):
        """Takes the operational_set and overwrites it with the set of nodes that results
        from applying the rule (this might or not include the initial set of nodes as
        well depending on the rule).

        :type operational_set: :py:class:`aiida.tools.graph.age_entities.Basket`
        :param operational_set: initital set of nodes to be overwritten by the rule.
        """


class QueryRule(Operation, metaclass=ABCMeta):
    """Parent class for every rule that implements a query.

    QueryRules take a generic QueryBuilder instance and a set of starting nodes and then
    perform successive iterations of that query, each one from the set of nodes that the
    previous one found. Depending on the class of rule used the final result will be
    either the whole set of nodes traversed (UpdateRule), or only the final set of nodes
    found in the last iteration of the query (ReplaceRule).
    """

    def __init__(self, querybuilder: orm.QueryBuilder, max_iterations=1, track_edges=False):
        """Initialization method

        :param querybuilder: an instance of the QueryBuilder class from which to take the
            procedure for traversal
        :param int max_iterations: the number of iterations to run this query on (must be
            a finite number for ReplaceRules)
        :param bool track_edges: whether to track which edges are traversed and store them
        """
        super().__init__(max_iterations, track_edges=track_edges)

        def get_spec_from_path(query_dict, idx):
            from aiida.orm.querybuilder import GROUP_ENTITY_TYPE_PREFIX

            if (
                query_dict['path'][idx]['entity_type'].startswith('node')
                or query_dict['path'][idx]['entity_type'].startswith('data')
                or query_dict['path'][idx]['entity_type'].startswith('process')
                or query_dict['path'][idx]['entity_type'] == ''
            ):
                result = 'nodes'
            elif query_dict['path'][idx]['entity_type'].startswith(GROUP_ENTITY_TYPE_PREFIX):
                result = 'groups'
            else:
                raise RuntimeError(f"not understood entity from ( {query_dict['path'][idx]['entity_type']} )")
            return result

        query_dict = querybuilder.as_dict()

        # Check if there is any projection:
        query_projections = query_dict['project']
        for projection_key in query_projections:
            if query_projections[projection_key] != []:
                raise ValueError(
                    'The input querybuilder must not have any projections.\n'
                    'Instead, it has the following:\n - Key: {}\n - Val: {}\n'.format(
                        projection_key, query_projections[projection_key]
                    )
                )
        for pathspec in query_dict['path']:
            if not pathspec['entity_type']:
                pathspec['entity_type'] = 'node.Node.'
        self._qbtemplate = deepcopy(querybuilder)
        query_dict = self._qbtemplate.as_dict()
        self._first_tag = query_dict['path'][0]['tag']
        self._last_tag = query_dict['path'][-1]['tag']
        self._querybuilder = None

        # All of these are set in _init_run:
        self._edge_label = None
        self._edge_keys = None
        self._entity_to_identifier = None

        self._entity_from = get_spec_from_path(query_dict, 0)
        self._entity_to = get_spec_from_path(query_dict, -1)
        self._accumulator_set = None

    def set_edge_keys(self, edge_keys):
        """Set the edge keys that are use to classify the edges during the run of this query.

        :param edge_keys:
            a list of projections on the edge itself, or a tuple that specifies
            (tag, project) if the projection is not on the edge

        Example: For node-to-node graph traversals, it is often convenient to save
        the information on the links::

            qb  = QueryBuilder().append(Node, tag='n1').append(Node, tag='n2')
            rule = RuleSequence(qb, track_edges=True)
            rule.set_edge_keys(['input_id', 'output_id', 'label', 'type'])

            # Now for UUIDS:
            qb  = QueryBuilder().append(Node, tag='n1').append(Node, tag='n2')
            rule = RuleSequence(qb, track_edges=True)
            rule.set_edge_keys([('n1','uuid'), ('n2','uuid'), 'label', 'type'])
        """
        self._edge_keys = edge_keys[:]

    def _init_run(self, operational_set):
        """Initialization Utility method

        This method initializes a run. It initializes the accumulator_set in order
        for it to only contain the operational_set, and to be of the same kind.
        This function modifies the its QueryBuilder instance to give the right results.

        :param operational_set: input with which to initialize the accumulator_set.
        """
        type_check(operational_set, Basket)
        if self._accumulator_set is not None:
            type_check(self._accumulator_set, Basket)
            self._accumulator_set.empty()
            self._accumulator_set += operational_set
        else:
            self._accumulator_set = operational_set.copy()

        # Copying qbtemplate so there's no problem if it is used again in a later run:
        query_dict = self._qbtemplate.as_dict()
        self._querybuilder = deepcopy(self._qbtemplate)

        self._entity_to_identifier = operational_set[self._entity_to].identifier

        # Now I add the necessary projections, which is the identifier of the
        # last entity of the QueryBuilder path:
        self._querybuilder.add_projection(self._last_tag, self._entity_to_identifier)

        if self._track_edges:
            # This requires additional projections and the edge_keys, which is a list of tuples (of length 2)
            # that stores the information what I need to project as well, as in (tag, projection)
            projections = defaultdict(list)
            self._edge_keys = []
            self._edge_label = query_dict['path'][-1]['edge_tag']

            # Need to get the edge_set: This is given by entity1_entity2. Here, the results needs to
            # be sorted somehow in order to ensure that the same key is used when entity_from and
            # entity_to are exchanged.
            edge_set = operational_set.dict['{}_{}'.format(*sorted((self._entity_from, self._entity_to)))]

            # Looping over the edge identifiers to figure out what I need to project and in which
            # order. The order is important! The rules:
            # r1 = Rule(QueryBuilder().append(Group).append(Node, with_group=Group) and
            # r2 = Rule(QueryBuilder().append(Node).append(Group, with_node=Node)
            # need still to save their results in the same order (i.e. group_id, node_id).
            # Therefore, I am sorting the edge_keys according to edge_identifiers specified in the edge_set
            for tag, projection in edge_set.edge_identifiers:
                if tag == 'edge':
                    actual_tag = self._edge_label
                elif tag == self._entity_from:
                    actual_tag = self._first_tag
                elif tag == self._entity_to:
                    actual_tag = self._last_tag
                else:
                    # For now I can only specify edge_identifiers as 'edge', ie. project on the edge
                    # itself, or by the entity_from, entity_to keyword, ie. groups or nodes.
                    # One could think of other keywords...
                    raise ValueError(f'This tag ({tag}) is not known')
                self._edge_keys.append((actual_tag, projection))
                projections[actual_tag].append(projection)

            # Telling the QB about the additional projections:
            for proj_tag, projectionlist in projections.items():
                try:
                    self._querybuilder.add_projection(proj_tag, projectionlist)
                except (TypeError, ValueError) as exc:
                    raise KeyError('The projection for the edge-identifier is invalid.\n') from exc

    def _load_results(self, target_set, operational_set):
        """Single application of the rules to the operational set

        :param target_set:
            where the new results will be loaded (it will be first emptied of all previous content).
            There is no returned value for this method.
        :param operational_set: where the results originate from (walkers)
        """
        primkeys = operational_set[self._entity_from].keyset
        target_set.empty()

        if primkeys:
            self._querybuilder.add_filter(
                self._first_tag, {operational_set[self._entity_from].identifier: {'in': primkeys}}
            )
            qres = self._querybuilder.dict()

            # These are the new results returned by the query
            target_set[self._entity_to].add_entities(
                [item[self._last_tag][self._entity_to_identifier] for item in qres]
            )

            if self._track_edges:
                # As in _init_run, I need the key for the edge_set
                edge_key = '{}_{}'.format(*sorted((self._entity_from, self._entity_to)))
                edge_set = operational_set.dict[edge_key]
                namedtuple_ = edge_set.edge_namedtuple

                target_set[edge_key].add_entities(
                    [namedtuple_(*(item[key1][key2] for (key1, key2) in self._edge_keys)) for item in qres]
                )

    def set_accumulator(self, accumulator_set):
        self._accumulator_set = accumulator_set

    def empty_accumulator(self):
        if self._accumulator_set is not None:
            self._accumulator_set.empty()

    # Pylint complains if this is not here, but should be removed asap
    def run(self, operational_set):
        pass


class UpdateRule(QueryRule):
    """The UpdateRule will accumulate every node visited and return it as a set of nodes
    (and thus, without duplication). It can be used requesting both a finite number
    of iterations or an infinite number of iterations (in which case it will stop once
    no new nodes are added to the accumulation set).
    """

    def run(self, operational_set):
        self._init_run(operational_set)
        self._iterations_done = 0
        new_results = operational_set.get_template()

        # The operational_set will be updated with the new_nodes that were not
        # already in the _acumulator_set, so that we are not querying from the
        # same nodes again and the cycle can end when no new nodes are found
        while operational_set and self._iterations_done < self._max_iterations:
            self._iterations_done += 1
            self._load_results(new_results, operational_set)
            operational_set = new_results - self._accumulator_set
            self._accumulator_set += new_results

        return self._accumulator_set.copy()


class ReplaceRule(QueryRule):
    """The ReplaceRule does not accumulate results, but just sets the operational_set to
    new results. Therefore it can only function using a finite number of iterations,
    since it does not keep track of which nodes where visited already (otherwise, if
    it was following a cycle, it would run indefinitely).
    """

    def __init__(self, querybuilder, max_iterations=1, track_edges=False):
        if max_iterations == np.inf:
            raise ValueError('You cannot have max_iterations to be infinitely large for replace rules')
        super().__init__(querybuilder, max_iterations=max_iterations, track_edges=track_edges)

    def run(self, operational_set):
        self._init_run(operational_set)
        self._iterations_done = 0
        new_results = operational_set.get_template()

        # The operational_set will be replaced by the new_nodes, even if these
        # were already visited previously.
        while operational_set and self._iterations_done < self._max_iterations:
            self._iterations_done += 1
            self._load_results(new_results, operational_set)
            operational_set = new_results

        return operational_set.copy()


class RuleSaveWalkers(Operation):
    """Save the Walkers:

    When initialized, this rule will save a pointer to an external stash variable.
    When run, this stash will be emptied and a given operational_set will be saved
    there instead.
    """

    def __init__(self, stash):
        """Initialization method

        :param stash: external variable in which to save the operational_set
        """
        self._stash = stash
        super().__init__(max_iterations=1, track_edges=True)

    def run(self, operational_set):
        self._stash.empty()
        self._stash += operational_set
        return operational_set


class RuleSetWalkers(Operation):
    """Set the Walkers:

    When initialized, this rule will save a pointer to an external stash variable.
    When run, the given operational_set will be emptied and the stash will be
    loaded in it.
    """

    def __init__(self, stash):
        """Initialization method

        :param stash: external variable from which to load into the operational_set
        """
        self._stash = stash
        super().__init__(max_iterations=1, track_edges=True)

    def run(self, operational_set):
        operational_set.empty()
        operational_set += self._stash
        return operational_set


class RuleSequence(Operation):
    """Rule for concatenation

    Rule Sequence is used to concatenate a series of rules together.
    Concatenating querybuilders in a single rule its not enough because
    one might want to stash results to perform two independent operations
    in the starting set instead of a second operation from the results of
    the first (see RuleSetWalkers and RuleSaveWalkers).
    """

    def __init__(self, rules, max_iterations=1):
        for rule in rules:
            if not isinstance(rule, Operation):
                raise TypeError('rule has to be an instance of Operation-subclass')
        self._rules = rules
        self._accumulator_set = None
        self._visits_set = None
        super().__init__(max_iterations, track_edges=False)

    def run(self, operational_set):
        type_check(operational_set, Basket)

        if self._accumulator_set is not None:
            type_check(self._accumulator_set, Basket)
            self._accumulator_set.empty()
            self._accumulator_set += operational_set
        else:
            self._accumulator_set = operational_set.copy()
        if self._visits_set is not None:
            type_check(self._visits_set, Basket)
            self._visits_set.empty()
            self._visits_set += operational_set
        else:
            self._visits_set = operational_set.copy()

        new_results = operational_set.get_template()
        self._iterations_done = 0
        while operational_set and self._iterations_done < self._max_iterations:
            self._iterations_done += 1
            new_results.empty()

            # I iterate the operational_set through all the rules:
            for _, rule in enumerate(self._rules):
                operational_set = rule.run(operational_set)
                new_results += operational_set
                self._visits_set += operational_set

            # I set the operational set to all results that have not been visited yet.
            operational_set = new_results - self._accumulator_set
            self._accumulator_set += new_results

        return self._visits_set.copy()

    def set_accumulator(self, accumulator_set):
        """Set the accumulator set"""
        self._accumulator_set = accumulator_set

    def empty_accumulator(self):
        """Empties the accumulator set"""
        if self._accumulator_set is not None:
            self._accumulator_set.empty()

    def set_visits(self, visits_set):
        """Set the visits set"""
        self._visits_set = visits_set

    def empty_visits(self):
        """Empties the visits set"""
        if self._visits_set is not None:
            self._visits_set.empty()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Class and decorators to generate processes out of simple python functions."""

from __future__ import annotations

import collections
import functools
import inspect
import itertools
import logging
import signal
import sys
import types
import typing as t
from typing import TYPE_CHECKING

import docstring_parser

from aiida.common.lang import override
from aiida.manage import get_manager
from aiida.orm import (
    Bool,
    CalcFunctionNode,
    Data,
    Dict,
    Float,
    Int,
    List,
    ProcessNode,
    Str,
    WorkFunctionNode,
    to_aiida_type,
)
from aiida.orm.utils.mixins import FunctionCalculationMixin

from .process import Process
from .process_spec import ProcessSpec

try:
    UnionType = types.UnionType
except AttributeError:
    # This type is not available for Python 3.9 and older
    UnionType = None  # type: ignore[assignment,misc]

try:
    from typing import ParamSpec
except ImportError:
    # Fallback for Python 3.9 and older
    from typing_extensions import ParamSpec  # type: ignore[assignment]

try:
    get_annotations = inspect.get_annotations
except AttributeError:
    # This is the backport for Python 3.9 and older
    from get_annotations import get_annotations  # type: ignore[no-redef]

if TYPE_CHECKING:
    from .exit_code import ExitCode

__all__ = ('calcfunction', 'workfunction', 'FunctionProcess')

LOGGER = logging.getLogger(__name__)

FunctionType = t.TypeVar('FunctionType', bound=t.Callable[..., t.Any])


def get_stack_size(size: int = 2) -> int:  # type: ignore[return]
    """Return the stack size for the caller's frame.

    This solution is taken from https://stackoverflow.com/questions/34115298/ as a more performant alternative to the
    naive ``len(inspect.stack())` solution. This implementation is about three orders of magnitude faster compared to
    the naive solution and it scales especially well for larger stacks, which will be usually the case for the usage
    of ``aiida-core``. However, it does use the internal ``_getframe`` of the ``sys`` standard library. It this ever
    were to stop working, simply switch to using ``len(inspect.stack())``.

    :param size: Hint for the expected stack size.
    :returns: The stack size for caller's frame.
    """
    frame = sys._getframe(size)
    try:
        for size in itertools.count(size, 8):
            frame = frame.f_back.f_back.f_back.f_back.f_back.f_back.f_back.f_back  # type: ignore[assignment,union-attr]
    except AttributeError:
        while frame:  # type: ignore[truthy-bool]
            frame = frame.f_back  # type: ignore[assignment]
            size += 1
        return size - 1  # type: ignore[unreachable]


P = ParamSpec('P')
R_co = t.TypeVar('R_co', covariant=True)
N = t.TypeVar('N', bound=ProcessNode)


class ProcessFunctionType(t.Protocol, t.Generic[P, R_co, N]):
    """Protocol for a decorated process function."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R_co: ...

    def run(self, *args: P.args, **kwargs: P.kwargs) -> R_co: ...

    def run_get_pk(self, *args: P.args, **kwargs: P.kwargs) -> tuple[dict[str, t.Any] | None, int]: ...

    def run_get_node(self, *args: P.args, **kwargs: P.kwargs) -> tuple[dict[str, t.Any] | None, N]: ...

    is_process_function: bool

    node_class: t.Type[N]

    process_class: t.Type[Process]

    recreate_from: t.Callable[[N], Process]

    spec: t.Callable[[], ProcessSpec]


def calcfunction(function: t.Callable[P, R_co]) -> ProcessFunctionType[P, R_co, CalcFunctionNode]:
    """A decorator to turn a standard python function into a calcfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the calcfunction
    >>> @calcfunction
    >>> def sum(a, b):
    >>>    return a + b
    >>> # Run it with some input
    >>> r = sum(Int(4), Int(5))
    >>> print(r)
    9
    >>> r.base.links.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<CalcFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.base.links.get_incoming().get_node_by_label('result').base.links.get_incoming().all_nodes()
    [4, 5]

    :param function: The function to decorate.
    :return: The decorated function.
    """
    return process_function(node_class=CalcFunctionNode)(function)  # type: ignore[arg-type]


def workfunction(function: t.Callable[P, R_co]) -> ProcessFunctionType[P, R_co, WorkFunctionNode]:
    """A decorator to turn a standard python function into a workfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the workfunction
    >>> @workfunction
    >>> def select(a, b):
    >>>    return a
    >>> # Run it with some input
    >>> r = select(Int(4), Int(5))
    >>> print(r)
    4
    >>> r.base.links.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<WorkFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.base.links.get_incoming().get_node_by_label('result').base.links.get_incoming().all_nodes()
    [4, 5]

    :param function: The function to decorate.
    :return: The decorated function.
    """
    return process_function(node_class=WorkFunctionNode)(function)  # type: ignore[arg-type]


def process_function(node_class: t.Type['ProcessNode']) -> t.Callable[[FunctionType], FunctionType]:
    """The base function decorator to create a FunctionProcess out of a normal python function.

    :param node_class: the ORM class to be used as the Node record for the FunctionProcess
    """

    def decorator(function: FunctionType) -> FunctionType:
        """Turn the decorated function into a FunctionProcess.

        :param callable function: the actual decorated function that the FunctionProcess represents
        :return callable: The decorated function.
        """
        process_class = FunctionProcess.build(function, node_class=node_class)

        def run_get_node(*args, **kwargs) -> tuple[dict[str, t.Any] | None, 'ProcessNode']:
            """Run the FunctionProcess with the supplied inputs in a local runner.

            :param args: input arguments to construct the FunctionProcess
            :param kwargs: input keyword arguments to construct the FunctionProcess
            :return: tuple of the outputs of the process and the process node
            """
            frame_delta = 1000
            frame_count = get_stack_size()
            stack_limit = sys.getrecursionlimit()
            LOGGER.info('Executing process function, current stack status: %d frames of %d', frame_count, stack_limit)

            # If the current frame count is more than 80% of the stack limit, or comes within 200 frames, increase the
            # stack limit by ``frame_delta``.
            if frame_count > min(0.8 * stack_limit, stack_limit - 200):
                LOGGER.warning(
                    'Current stack contains %d frames which is close to the limit of %d. Increasing the limit by %d',
                    frame_count,
                    stack_limit,
                    frame_delta,
                )
                sys.setrecursionlimit(stack_limit + frame_delta)

            manager = get_manager()
            runner = manager.get_runner()
            inputs = process_class.create_inputs(*args, **kwargs)

            # Remove all the known inputs from the kwargs
            for port in process_class.spec().inputs:
                kwargs.pop(port, None)

            # If any kwargs remain, the spec should be dynamic, so we raise if it isn't
            if kwargs and not process_class.spec().inputs.dynamic:
                raise ValueError(f'{function.__name__} does not support these kwargs: {kwargs.keys()}')

            process = process_class(inputs=inputs, runner=runner)

            # Only add handlers for interrupt signal to kill the process if we are in a local and not a daemon runner.
            # Without this check, running process functions in a daemon worker would be killed if the daemon is shutdown
            current_runner = manager.get_runner()
            original_handler = None
            kill_signal = signal.SIGINT

            if not current_runner.is_daemon_runner:

                def kill_process(_num, _frame):
                    """Send the kill signal to the process in the current scope."""
                    LOGGER.critical('runner received interrupt, killing process %s', process.pid)
                    result = process.kill(msg='Process was killed because the runner received an interrupt')
                    return result

                # Store the current handler on the signal such that it can be restored after process has terminated
                original_handler = signal.getsignal(kill_signal)
                signal.signal(kill_signal, kill_process)

            try:
                result = process.execute()
            finally:
                # If the `original_handler` is set, that means the `kill_process` was bound, which needs to be reset
                if original_handler:
                    signal.signal(signal.SIGINT, original_handler)

            store_provenance = inputs.get('metadata', {}).get('store_provenance', True)
            if not store_provenance:
                process.node._storable = False
                process.node._unstorable_message = 'cannot store node because it was run with `store_provenance=False`'

            return result, process.node

        def run_get_pk(*args, **kwargs) -> tuple[dict[str, t.Any] | None, int]:
            """Recreate the `run_get_pk` utility launcher.

            :param args: input arguments to construct the FunctionProcess
            :param kwargs: input keyword arguments to construct the FunctionProcess
            :return: tuple of the outputs of the process and the process node pk

            """
            result, node = run_get_node(*args, **kwargs)
            assert node.pk is not None
            return result, node.pk

        @functools.wraps(function)
        def decorated_function(*args, **kwargs):
            """This wrapper function is the actual function that is called."""
            result, _ = run_get_node(*args, **kwargs)
            return result

        decorated_function.run = decorated_function  # type: ignore[attr-defined]
        decorated_function.run_get_pk = run_get_pk  # type: ignore[attr-defined]
        decorated_function.run_get_node = run_get_node  # type: ignore[attr-defined]
        decorated_function.is_process_function = True  # type: ignore[attr-defined]
        decorated_function.node_class = node_class  # type: ignore[attr-defined]
        decorated_function.process_class = process_class  # type: ignore[attr-defined]
        decorated_function.recreate_from = process_class.recreate_from  # type: ignore[attr-defined]
        decorated_function.spec = process_class.spec  # type: ignore[attr-defined]

        return decorated_function  # type: ignore[return-value]

    return decorator


def infer_valid_type_from_type_annotation(annotation: t.Any) -> tuple[t.Any, ...]:
    """Infer the value for the ``valid_type`` of an input port from the given function argument annotation.

    :param annotation: The annotation of a function argument as returned by ``inspect.get_annotation``.
    :returns: A tuple of valid types. If no valid types were defined or they could not be successfully parsed, an empty
        tuple is returned.
    """

    def get_type_from_annotation(annotation):
        valid_type_map = {
            bool: Bool,
            dict: Dict,
            t.Dict: Dict,
            float: Float,
            int: Int,
            list: List,
            t.List: List,
            str: Str,
        }

        if inspect.isclass(annotation) and issubclass(annotation, Data):
            return annotation

        return valid_type_map.get(annotation)

    inferred_valid_type: tuple[t.Any, ...] = ()

    if inspect.isclass(annotation):
        inferred_valid_type = (get_type_from_annotation(annotation),)
    elif t.get_origin(annotation) is t.Union or t.get_origin(annotation) is UnionType:
        inferred_valid_type = tuple(get_type_from_annotation(valid_type) for valid_type in t.get_args(annotation))
    elif t.get_origin(annotation) is t.Optional:
        inferred_valid_type = (t.get_args(annotation),)

    return tuple(valid_type for valid_type in inferred_valid_type if valid_type is not None)


class FunctionProcess(Process):
    """Function process class used for turning functions into a Process"""

    _func_args: t.Sequence[str] = ()
    _var_positional: str | None = None
    _var_keyword: str | None = None

    @staticmethod
    def _func(*_args, **_kwargs) -> dict:
        """This is used internally to store the actual function that is being
        wrapped and will be replaced by the build method.
        """
        return {}

    @staticmethod
    def build(func: FunctionType, node_class: t.Type['ProcessNode']) -> t.Type['FunctionProcess']:
        """Build a Process from the given function.

        All function arguments will be assigned as process inputs. If keyword arguments are specified then
        these will also become inputs.

        :param func: The function to build a process from
        :param node_class: Provide a custom node class to be used, has to be constructable with no arguments. It has to
            be a sub class of `ProcessNode` and the mixin :class:`~aiida.orm.utils.mixins.FunctionCalculationMixin`.

        :return: A Process class that represents the function

        """
        if (
            not issubclass(node_class, ProcessNode)  # type: ignore[redundant-expr]
            or not issubclass(node_class, FunctionCalculationMixin)  # type: ignore[unreachable]
        ):
            raise TypeError('the node_class should be a sub class of `ProcessNode` and `FunctionCalculationMixin`')

        signature = inspect.signature(func)  # type: ignore[unreachable]

        args: list[str] = []
        var_positional: str | None = None
        var_keyword: str | None = None

        try:
            annotations = get_annotations(func, eval_str=True)
        except Exception as exception:
            # Since we are running with ``eval_str=True`` to unstringize the annotations, the call can except if the
            # annotations are incorrect. In this case we simply want to log a warning and continue with type inference.
            LOGGER.warning(f'function `{func.__name__}` has invalid type hints: {exception}')
            annotations = {}

        try:
            parsed_docstring = docstring_parser.parse(func.__doc__)
        except Exception as exception:
            LOGGER.warning(f'function `{func.__name__}` has a docstring that could not be parsed: {exception}')
            param_help_string = {}
            namespace_help_string = None
        else:
            param_help_string = {param.arg_name: param.description for param in parsed_docstring.params}
            namespace_help_string = parsed_docstring.short_description if parsed_docstring.short_description else ''
            if parsed_docstring.long_description is not None:
                namespace_help_string += f'\n\n{parsed_docstring.long_description}'

        for key, parameter in signature.parameters.items():
            if parameter.kind in [parameter.POSITIONAL_ONLY, parameter.POSITIONAL_OR_KEYWORD, parameter.KEYWORD_ONLY]:
                args.append(key)

            if parameter.kind is parameter.VAR_POSITIONAL:
                var_positional = key

            if parameter.kind is parameter.VAR_KEYWORD:
                var_keyword = key

        def define(cls, spec):
            """Define the spec dynamically"""
            from plumpy.ports import UNSPECIFIED

            super().define(spec)

            for parameter in signature.parameters.values():
                if parameter.kind in [parameter.VAR_POSITIONAL, parameter.VAR_KEYWORD]:
                    continue

                annotation = annotations.get(parameter.name)
                valid_type = infer_valid_type_from_type_annotation(annotation) or (Data,)
                help_string = param_help_string.get(parameter.name, None)

                default = parameter.default if parameter.default is not parameter.empty else UNSPECIFIED

                # If the keyword was already specified, simply override the default
                if spec.has_input(parameter.name):
                    spec.inputs[parameter.name].default = default
                    continue

                # If the default is ``None`` make sure that the port also accepts a ``NoneType``. Note that we cannot
                # use ``None`` because the validation will call ``isinstance`` which does not work when passing ``None``
                # but it does work with ``NoneType`` which is returned by calling ``type(None)``.
                if default is None:
                    valid_type += (type(None),)

                # If a default is defined and it is not a ``Data`` instance it should be serialized, but this should be
                # done lazily using a lambda, just as any port defaults should not define node instances directly as is
                # also checked by the ``spec.input`` call.
                if (
                    default is not None
                    and default != UNSPECIFIED
                    and not isinstance(default, Data)
                    and not callable(default)
                ):

                    def indirect_default(value=default):
                        return to_aiida_type(value)
                else:
                    indirect_default = default

                spec.input(
                    parameter.name,
                    valid_type=valid_type,
                    default=indirect_default,
                    serializer=to_aiida_type,
                    help=help_string,
                )

            # Set defaults for label and description based on function name and docstring, if not explicitly defined
            port_label = spec.inputs['metadata']['label']

            if not port_label.has_default():
                port_label.default = func.__name__

            spec.inputs.help = namespace_help_string

            # If the function supports varargs or kwargs then allow dynamic inputs, otherwise disallow
            spec.inputs.dynamic = var_positional is not None or var_keyword is not None

            # Function processes must have a dynamic output namespace since we do not know beforehand what outputs
            # will be returned and the valid types for the value should be `Data` nodes as well as a dictionary because
            # the output namespace can be nested.
            spec.outputs.valid_type = (Data, dict)

        return type(
            func.__qualname__,
            (FunctionProcess,),
            {
                '__module__': func.__module__,
                '__name__': func.__name__,
                '__qualname__': func.__qualname__,
                '_func': staticmethod(func),
                Process.define.__name__: classmethod(define),
                '_func_args': args,
                '_var_positional': var_positional,
                '_var_keyword': var_keyword,
                '_node_class': node_class,
            },
        )

    @classmethod
    def validate_inputs(cls, *args: t.Any, **kwargs: t.Any) -> None:
        """Validate the positional and keyword arguments passed in the function call.

        :raises TypeError: if more positional arguments are passed than the function defines
        """
        nargs = len(args)
        nparameters = len(cls._func_args)

        # If the number of positional arguments passed is larger than the number of explicitly defined parameters in the
        # signature and the function does not support variable positional arguments the inputs are invalid and we raise.
        # If we don't, some of the passed arguments, intended to be positional arguments, will be misinterpreted as
        # keyword arguments, but they won't have an explicit name to use for the link label, causing the input link to
        # be completely lost. If the function supports variadic arguments, however, additional args should be accepted.
        if nargs > nparameters and cls._var_positional is None:
            name = cls._func.__name__
            raise TypeError(f'{name}() takes {nparameters} positional arguments but {nargs} were given')

    @classmethod
    def create_inputs(cls, *args: t.Any, **kwargs: t.Any) -> dict[str, t.Any]:
        """Create the input dictionary for the ``FunctionProcess``."""
        cls.validate_inputs(*args, **kwargs)

        # The complete input dictionary consists of the keyword arguments...
        inputs = dict(kwargs or {})
        arguments = list(args)

        # ... and then the positional arguments need be added. Arguments with an explicit positional argument simply use
        # the argument name as the key in the dictionary explicit label. For variable positional arguments, the label is
        # constructed based on the name used for the variable positional arguments, which is stored in the
        # ``cls._var_positional`` class attribute and is suffixed with the index. If the auto-generated label would
        # overlap with an existing label (i.e. from one of the keyword arguments) an exception is raised.
        for name, parameter in inspect.signature(cls._func).parameters.items():
            if parameter.kind in [parameter.POSITIONAL_ONLY, parameter.POSITIONAL_OR_KEYWORD]:
                try:
                    inputs[name] = arguments.pop(0)
                except IndexError:
                    pass
            elif name == cls._var_positional and parameter.kind is parameter.VAR_POSITIONAL:
                for index, arg in enumerate(arguments):
                    label = f'{cls._var_positional}_{index}'
                    if label in inputs:
                        raise RuntimeError(
                            f'variadic argument with index `{index}` would get the label `{label}` but this is already '
                            'in use by another function argument with the exact same name. To avoid this error, please '
                            f'change the name of argument `{label}` to something else.'
                        )
                    inputs[label] = arg

        return inputs

    @classmethod
    def get_or_create_db_record(cls) -> 'ProcessNode':
        return cls._node_class()

    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('enable_persistence', False):
            raise RuntimeError('Cannot persist a function process')
        super().__init__(enable_persistence=False, *args, **kwargs)  # type: ignore[misc]

    @property
    def process_class(self) -> t.Callable[..., t.Any]:
        """Return the class that represents this Process, for the FunctionProcess this is the function itself.

        For a standard Process or sub class of Process, this is the class itself. However, for legacy reasons,
        the Process class is a wrapper around another class. This function returns that original class, i.e. the
        class that really represents what was being executed.

        :return: A Process class that represents the function

        """
        return self._func

    def execute(self) -> dict[str, t.Any] | None:
        """Execute the process."""
        result = super().execute()

        # FunctionProcesses can return a single value as output, and not a dictionary, so we should also return that
        if result and len(result) == 1 and self.SINGLE_OUTPUT_LINKNAME in result:
            return result[self.SINGLE_OUTPUT_LINKNAME]

        return result

    @override
    def _setup_db_record(self) -> None:
        """Set up the database record for the process."""
        super()._setup_db_record()
        self.node.store_source_info(self._func)

    @override
    def run(self) -> 'ExitCode' | None:
        """Run the process."""
        from .exit_code import ExitCode

        # The following conditional is required for the caching to properly work. Even if the source node has a process
        # state of `Finished` the cached process will still enter the running state. The process state will have then
        # been overridden by the engine to `Running` so we cannot check that, but if the `exit_status` is anything other
        # than `None`, it should mean this node was taken from the cache, so the process should not be rerun.
        if self.node.exit_status is not None:
            return ExitCode(self.node.exit_status, self.node.exit_message)

        # Now the original functions arguments need to be reconstructed from the inputs to the process, as they were
        # passed to the original function call. To do so, all positional parameters are popped from the inputs
        # dictionary and added to the positional arguments list.
        args = []
        kwargs: dict[str, Data] = {}
        inputs = dict(self.inputs or {})

        for name, parameter in inspect.signature(self._func).parameters.items():
            if parameter.kind in [parameter.POSITIONAL_ONLY, parameter.POSITIONAL_OR_KEYWORD]:
                args.append(inputs.pop(name))
            elif parameter.kind is parameter.VAR_POSITIONAL:
                for key in [key for key in inputs.keys() if key.startswith(f'{name}_')]:
                    args.append(inputs.pop(key))

        # Any inputs that correspond to metadata ports were not part of the original function signature but were added
        # by the process function decorator, so these have to be removed.
        for key in [key for key, port in self.spec().inputs.items() if port.is_metadata]:
            inputs.pop(key, None)

        # The remaining inputs have to be keyword arguments.
        kwargs.update(**inputs)

        result = self._func(*args, **kwargs)

        if result is None or isinstance(result, ExitCode):  # type: ignore[redundant-expr]
            return result  # type: ignore[unreachable]

        if isinstance(result, Data):  # type: ignore[unreachable]
            self.out(self.SINGLE_OUTPUT_LINKNAME, result)  # type: ignore[unreachable]
        elif isinstance(result, collections.abc.Mapping):
            for name, value in result.items():
                self.out(name, value)
        else:
            raise TypeError(
                "Function process returned an output with unsupported type '{}'\n"
                'Must be a Data type or a mapping of {{string: Data}}'.format(result.__class__)
            )

        return ExitCode()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility for performing schema migrations, via reflection of the current database."""

from __future__ import annotations

import alembic
from sqlalchemy import inspect


class ReflectMigrations:
    """Perform schema migrations, via reflection of the current database.

    In django, it is not possible to explicitly specify constraints/indexes and their names,
    instead they are implicitly created by internal "auto-generation" code
    (as opposed to sqlalchemy, where one can explicitly specify the names).
    For a specific django version, this auto-generation code is deterministic,
    however, over time it has changed.
    So is not possible to know declaratively exactly what constraints/indexes are present on a users database,
    withtout knowing the exact django version that created it (and run migrations).
    Therefore, we need to reflect the database's schema, to determine what is present on the database,
    to know what to drop.
    """

    def __init__(self, op: alembic.op) -> None:
        self.op = op
        # note, we only want to instatiate the inspector once, since it caches reflection calls to the database
        self.inspector = inspect(op.get_bind())

    def reset_cache(self) -> None:
        """Reset the inspector cache."""
        self.inspector = inspect(self.op.get_bind())

    def drop_all_unique_constraints(self, table_name: str) -> None:
        """Drop all unique constraints set for this table."""
        for constraint in self.inspector.get_unique_constraints(table_name):
            self.op.drop_constraint(constraint['name'], table_name, type_='unique')

    def drop_unique_constraints(self, table_name: str, column_names: list[str]) -> None:
        """Drop all unique constraints set for this column name group."""
        column_set = set(column_names)
        for constraint in self.inspector.get_unique_constraints(table_name):
            if set(constraint['column_names']) == column_set:
                self.op.drop_constraint(constraint['name'], table_name, type_='unique')

    def drop_all_indexes(self, table_name: str, unique: bool = False) -> None:
        """Drop all non-unique indexes set for this table."""
        for index in self.inspector.get_indexes(table_name):
            if index['unique'] is unique:
                self.op.drop_index(index['name'], table_name)

    def drop_indexes(self, table_name: str, column: str | list[str], unique: bool = False) -> None:
        """Drop all indexes set for this column name group."""
        if isinstance(column, str):
            column = [column]
        column_set = set(column)
        for index in self.inspector.get_indexes(table_name):
            if (index['unique'] is unique) and (set(index['column_names']) == column_set):
                self.op.drop_index(index['name'], table_name)

    def drop_all_foreign_keys(self, table_name: str) -> None:
        """Drop all foreign keys set for this table."""
        for constraint in self.inspector.get_foreign_keys(table_name):
            self.op.drop_constraint(constraint['name'], table_name, type_='foreignkey')

    def drop_foreign_keys(self, table_name: str, columns: list[str], ref_tbl: str, ref_columns: list[str]) -> None:
        """Drop all foreign keys set for this column name group and referring column set."""
        column_set = set(columns)
        ref_column_set = set(ref_columns)
        for constraint in self.inspector.get_foreign_keys(table_name):
            if constraint['referred_table'] != ref_tbl:
                continue
            if set(constraint['referred_columns']) != ref_column_set:
                continue
            if set(constraint['constrained_columns']) == column_set:
                self.op.drop_constraint(constraint['name'], table_name, type_='foreignkey')

    def replace_index(self, label: str, table_name: str, column: str, unique: bool = False) -> None:
        """Create index, dropping any existing index with the same table+columns."""
        self.drop_indexes(table_name, column, unique)
        self.op.create_index(label, table_name, column, unique=unique)

    def replace_unique_constraint(self, label: str, table_name: str, columns: list[str]) -> None:
        """Create unique constraint, dropping any existing unique constraint with the same table+columns."""
        self.drop_unique_constraints(table_name, columns)
        self.op.create_unique_constraint(label, table_name, columns)

    def replace_foreign_key(
        self, label: str, table_name: str, columns: list[str], ref_tbl: str, ref_columns: list[str], **kwargs
    ) -> None:
        """Create foreign key, dropping any existing foreign key with the same constraints."""
        self.drop_foreign_keys(table_name, columns, ref_tbl, ref_columns)
        self.op.create_foreign_key(label, table_name, ref_tbl, columns, ref_columns, **kwargs)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the SqlAlchemy backend implementation of the `AuthInfo` ORM class."""

from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.orm.implementation.authinfos import BackendAuthInfo, BackendAuthInfoCollection
from aiida.storage.psql_dos.models.authinfo import DbAuthInfo

from . import computers, entities, users, utils


class SqlaAuthInfo(entities.SqlaModelEntity[DbAuthInfo], BackendAuthInfo):
    """SqlAlchemy backend implementation for the `AuthInfo` ORM class."""

    MODEL_CLASS = DbAuthInfo
    USER_CLASS = users.SqlaUser
    COMPUTER_CLASS = computers.SqlaComputer

    def __init__(self, backend, computer, user):
        """Construct a new instance.

        :param computer: a :class:`aiida.orm.implementation.computers.BackendComputer` instance
        :param user: a :class:`aiida.orm.implementation.users.BackendUser` instance
        :return: an :class:`aiida.orm.implementation.authinfos.BackendAuthInfo` instance
        """
        super().__init__(backend)
        type_check(user, self.USER_CLASS)
        type_check(computer, self.COMPUTER_CLASS)
        self._model = utils.ModelWrapper(
            self.MODEL_CLASS(dbcomputer=computer.bare_model, aiidauser=user.bare_model), backend
        )

    @property
    def id(self):
        return self.model.id

    @property
    def is_stored(self) -> bool:
        """Return whether the entity is stored.

        :return: True if stored, False otherwise
        """
        return self.model.is_saved()

    @property
    def enabled(self) -> bool:
        """Return whether this instance is enabled.

        :return: boolean, True if enabled, False otherwise
        """
        return self.model.enabled

    @enabled.setter
    def enabled(self, enabled):
        """Set the enabled state

        :param enabled: boolean, True to enable the instance, False to disable it
        """
        self.model.enabled = enabled

    @property
    def computer(self):
        """Return the computer associated with this instance.

        :return: :class:`aiida.orm.implementation.computers.BackendComputer`
        """
        return self.backend.computers.ENTITY_CLASS.from_dbmodel(self.model.dbcomputer, self.backend)

    @property
    def user(self):
        """Return the user associated with this instance.

        :return: :class:`aiida.orm.implementation.users.BackendUser`
        """
        return self.backend.users.ENTITY_CLASS.from_dbmodel(self.model.aiidauser, self.backend)

    def get_auth_params(self):
        """Return the dictionary of authentication parameters

        :return: a dictionary with authentication parameters
        """
        return self.model.auth_params

    def set_auth_params(self, auth_params):
        """Set the dictionary of authentication parameters

        :param auth_params: a dictionary with authentication parameters
        """
        self.model.auth_params = auth_params

    def get_metadata(self):
        """Return the dictionary of metadata

        :return: a dictionary with metadata
        """
        return self.model._metadata

    def set_metadata(self, metadata):
        """Set the dictionary of metadata

        :param metadata: a dictionary with metadata
        """
        self.model._metadata = metadata


class SqlaAuthInfoCollection(BackendAuthInfoCollection):
    """The collection of SqlAlchemy backend `AuthInfo` entries."""

    ENTITY_CLASS = SqlaAuthInfo

    def delete(self, pk):
        """Delete an entry from the collection.

        :param pk: the pk of the entry to delete
        """
        from sqlalchemy.orm.exc import NoResultFound

        session = self.backend.get_session()

        try:
            row = session.query(self.ENTITY_CLASS.MODEL_CLASS).filter_by(id=pk).one()
            session.delete(row)
            session.commit()
        except NoResultFound:
            raise exceptions.NotExistent(f'AuthInfo<{pk}> does not exist')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration after the `Group` class became pluginnable and so the group `type_string` changed.

Revision ID: django_0044
Revises: django_0043

"""

import sqlalchemy as sa
from alembic import op

revision = 'django_0044'
down_revision = 'django_0043'
branch_labels = None
depends_on = None

forward_sql = [
    """UPDATE db_dbgroup SET type_string = 'core' WHERE type_string = 'user';""",
    """UPDATE db_dbgroup SET type_string = 'core.upf' WHERE type_string = 'data.upf';""",
    """UPDATE db_dbgroup SET type_string = 'core.import' WHERE type_string = 'auto.import';""",
    """UPDATE db_dbgroup SET type_string = 'core.auto' WHERE type_string = 'auto.run';""",
]


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()
    statement = sa.text('\n'.join(forward_sql))
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0044.')



"""Calculations provided by aiida_diff tutorial plugin.

Register calculations via the "aiida.calculations" entry point in the pyproject.toml file.
"""

from aiida.common import datastructures
from aiida.engine import CalcJob
from aiida.orm import SinglefileData


class DiffCalculation(CalcJob):
    """AiiDA calculation plugin wrapping the diff executable.

    Simple AiiDA plugin wrapper for 'diffing' two files.
    """

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        super(DiffCalculation, cls).define(spec)

        # new ports
        spec.input('file1', valid_type=SinglefileData, help='First file to be compared.')
        spec.input('file2', valid_type=SinglefileData, help='Second file to be compared.')
        spec.output('diff', valid_type=SinglefileData, help='diff between file1 and file2.')

        spec.input('metadata.options.output_filename', valid_type=str, default='patch.diff')
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }
        spec.inputs['metadata']['options']['parser_name'].default = 'diff-tutorial'

        spec.exit_code(
            300, 'ERROR_MISSING_OUTPUT_FILES', message='Calculation did not produce all expected output files.'
        )

    def prepare_for_submission(self, folder):
        """Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files needed by
            the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        codeinfo = datastructures.CodeInfo()
        codeinfo.cmdline_params = [self.inputs.file1.filename, self.inputs.file2.filename]
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = [
            (self.inputs.file1.uuid, self.inputs.file1.filename, self.inputs.file1.filename),
            (self.inputs.file2.uuid, self.inputs.file2.filename, self.inputs.file2.filename),
        ]
        calcinfo.retrieve_list = [self.metadata.options.output_filename]

        return calcinfo



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Seal any process nodes that have not yet been sealed but should.

This should have been accomplished by the last step in the previous migration, but because the WHERE clause was
incorrect, not all nodes that should have been targeted were included. The problem is with the statement:

    attributes->>'process_state' NOT IN ('created', 'running', 'waiting')

The problem here is that this will yield `False` if the attribute `process_state` does not even exist. This will be the
case for legacy calculations like `InlineCalculation` nodes. Their node type was already migrated in `0020` but most of
them will be unsealed.

This is identical to migration 7b38a9e783e7

Revision ID: django_0041
Revises: django_0040

"""

import sqlalchemy as sa
from alembic import op

revision = 'django_0041'
down_revision = 'django_0040'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    statement = sa.text(
        """
        UPDATE db_dbnode
        SET attributes = jsonb_set(attributes, '{"sealed"}', to_jsonb(True))
        WHERE
            node_type LIKE 'process.%' AND
            NOT attributes ? 'sealed' AND
            NOT (
                attributes ? 'process_state' AND
                attributes->>'process_state' IN ('created', 'running', 'waiting')
            );
        -- Set `sealed=True` for process nodes that do not yet have a `sealed` attribute AND are not in an active state
        -- It is important to check that `process_state` exists at all before doing the IN check.
        """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0041.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data core.remote` command."""

import stat

import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params import arguments, types
from aiida.cmdline.utils import echo


@verdi_data.group('core.remote')
def remote():
    """Manipulate RemoteData objects (reference to remote folders).

    A RemoteData can be thought as a "symbolic link" to a folder on one of the
    Computers set up in AiiDA (e.g. where a CalcJob will run).
    This folder is called "remote" in the sense that it is on a Computer and
    not in the AiiDA repository. Note, however, that the "remote" computer
    could also be "localhost".
    """


@remote.command('ls')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.remote',)))
@click.option('-l', '--long', 'ls_long', is_flag=True, default=False, help='Display also file metadata.')
@click.option('-p', '--path', type=click.STRING, default='.', help='The folder to list.')
def remote_ls(ls_long, path, datum):
    """List content of a (sub)directory in a RemoteData object."""
    import datetime

    try:
        content = datum.listdir_withattributes(path=path)
    except OSError as err:
        echo.echo_critical(f'Unable to access the remote folder or file, check if it exists.\nOriginal error: {err!s}')
    for metadata in content:
        if ls_long:
            mtime = datetime.datetime.fromtimestamp(metadata['attributes'].st_mtime)
            pre_line = '{} {:10}  {}  '.format(
                stat.filemode(metadata['attributes'].st_mode),
                metadata['attributes'].st_size,
                mtime.strftime('%d %b %Y %H:%M'),
            )
            echo.echo(pre_line, nl=False)
        if metadata['isdir']:
            echo.echo(metadata['name'], fg=echo.COLORS['info'])
        else:
            echo.echo(metadata['name'])


@remote.command('cat')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.remote',)))
@click.argument('path', type=click.STRING)
def remote_cat(datum, path):
    """Show content of a file in a RemoteData object."""
    import os
    import sys
    import tempfile

    try:
        with tempfile.NamedTemporaryFile(delete=False) as tmpf:
            tmpf.close()
            datum.getfile(path, tmpf.name)
            with open(tmpf.name, encoding='utf8') as fhandle:
                sys.stdout.write(fhandle.read())
    except OSError as err:
        echo.echo_critical(f'{err.errno}: {err!s}')

    try:
        os.remove(tmpf.name)
    except OSError:
        # If you cannot delete, ignore (maybe I didn't manage to create it in the first place
        pass


@remote.command('show')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.remote',)))
def remote_show(datum):
    """Show information for a RemoteData object."""
    echo.echo(f'- Remote computer name: {datum.computer.label}')
    echo.echo(f'- Remote folder full path: {datum.get_remote_path()}')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Provides all parameter types."""

# AUTO-GENERATED

# fmt: off

from .calculation import *
from .choice import *
from .code import *
from .computer import *
from .config import *
from .data import *
from .group import *
from .identifier import *
from .multiple import *
from .node import *
from .path import *
from .plugin import *
from .process import *
from .profile import *
from .strings import *
from .user import *
from .workflow import *

__all__ = (
    'AbsolutePathParamType',
    'CalculationParamType',
    'CodeParamType',
    'ComputerParamType',
    'ConfigOptionParamType',
    'DataParamType',
    'EmailType',
    'EntryPointType',
    'FileOrUrl',
    'GroupParamType',
    'HostnameType',
    'IdentifierParamType',
    'LabelStringType',
    'LazyChoice',
    'MpirunCommandParamType',
    'MultipleValueParamType',
    'NodeParamType',
    'NonEmptyStringParamType',
    'PathOrUrl',
    'PluginParamType',
    'ProcessParamType',
    'ProfileParamType',
    'ShebangParamType',
    'UserParamType',
    'WorkflowParamType',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Add a uniqueness constraint on `db_dbnode.uuid`.

Revision ID: django_0014
Revises: django_0013

"""

from alembic import op

revision = 'django_0014'
down_revision = 'django_0013'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    from aiida.storage.psql_dos.migrations.utils.duplicate_uuids import verify_uuid_uniqueness

    verify_uuid_uniqueness('db_dbnode', op.get_bind())
    op.create_unique_constraint('db_dbnode_uuid_62e0bf98_uniq', 'db_dbnode', ['uuid'])
    op.drop_index('db_dbnode_uuid_62e0bf98', table_name='db_dbnode')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0014.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Various utils that should be used during migrations and migrations tests because the AiiDA ORM cannot be used."""

import datetime
import functools
import io
import json
import os
import pathlib
import re
from typing import Dict, Iterable, List, Optional, Union

import numpy
from disk_objectstore import Container
from disk_objectstore.utils import LazyOpener

from aiida.common import exceptions
from aiida.repository.backend import AbstractRepositoryBackend
from aiida.repository.common import File, FileType
from aiida.repository.repository import Repository
from aiida.storage.psql_dos.backend import get_filepath_container

ISOFORMAT_DATETIME_REGEX = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+(\+\d{2}:\d{2})?$')
REGEX_SHARD_SUB_LEVEL = re.compile(r'^[0-9a-f]{2}$')
REGEX_SHARD_FINAL_LEVEL = re.compile(r'^[0-9a-f-]{32}$')


class LazyFile(File):
    """Subclass of `File` where `key` also allows `LazyOpener` in addition to a string.

    This subclass is necessary because the migration will be storing instances of `LazyOpener` as the `key` which should
    normally only be a string. This subclass updates the `key` type check to allow this.
    """

    def __init__(
        self,
        name: str = '',
        file_type: FileType = FileType.DIRECTORY,
        key: Union[str, None, LazyOpener] = None,
        objects: Optional[Dict[str, 'File']] = None,
    ):
        if not isinstance(name, str):
            raise TypeError('name should be a string.')

        if not isinstance(file_type, FileType):
            raise TypeError('file_type should be an instance of `FileType`.')

        if key is not None and not isinstance(key, (str, LazyOpener)):
            raise TypeError('key should be `None` or a string.')

        if objects is not None and any(not isinstance(obj, self.__class__) for obj in objects.values()):
            raise TypeError('objects should be `None` or a dictionary of `File` instances.')

        if file_type == FileType.DIRECTORY and key is not None:
            raise ValueError('an object of type `FileType.DIRECTORY` cannot define a key.')

        if file_type == FileType.FILE and objects is not None:
            raise ValueError('an object of type `FileType.FILE` cannot define any objects.')

        self._name = name
        self._file_type = file_type
        self._key = key
        self._objects = objects or {}


class MigrationRepository(Repository):
    """Subclass of `Repository` that uses `LazyFile` instead of `File` as its file class."""

    _file_cls = LazyFile


class NoopRepositoryBackend(AbstractRepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` where all write operations are no-ops.

    This repository backend is used to use the ``Repository`` interface to build repository metadata but instead of
    actually writing the content of the current repository to disk elsewhere, it will simply open a lazy file opener.
    In a subsequent step, all these streams are passed to the new Disk Object Store that will write their content
    directly to pack files for optimal efficiency.
    """

    @property
    def uuid(self) -> Optional[str]:
        """Return the unique identifier of the repository.

        .. note:: A sandbox folder does not have the concept of a unique identifier and so always returns ``None``.
        """
        return None

    @property
    def key_format(self) -> Optional[str]:
        return None

    def initialise(self, **kwargs) -> None:
        raise NotImplementedError()

    @property
    def is_initialised(self) -> bool:
        return True

    def erase(self):
        raise NotImplementedError()

    def _put_object_from_filelike(self, handle: io.BufferedIOBase) -> str:
        return LazyOpener(handle.name)

    def has_objects(self, keys: List[str]) -> List[bool]:
        raise NotImplementedError()

    def delete_objects(self, keys: List[str]) -> None:
        raise NotImplementedError()

    def list_objects(self) -> Iterable[str]:
        raise NotImplementedError()

    def iter_object_streams(self, keys: List[str]):
        raise NotImplementedError()

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        raise NotImplementedError

    def get_info(self, detailed: bool = False, **kwargs) -> dict:
        raise NotImplementedError


def migrate_legacy_repository(profile, shard=None):
    """Migrate the legacy file repository to the new disk object store and return mapping of repository metadata.

    .. warning:: this method assumes that the new disk object store container has been initialized.

    The format of the return value will be a dictionary where the keys are the UUIDs of the nodes whose repository
    folder has contents have been migrated to the disk object store. The values are the repository metadata that contain
    the keys for the generated files with which the files in the disk object store can be retrieved. The format of the
    repository metadata follows exactly that of what is generated normally by the ORM.

    This implementation consciously uses the ``Repository`` interface in order to not have to rewrite the logic that
    builds the nested repository metadata based on the contents of a folder on disk. The advantage is that in this way
    it is guarantee that the exact same repository metadata is generated as it would have during normal operation.
    However, if the ``Repository`` interface or its implementation ever changes, it is possible that this solution will
    have to be adapted and the significant parts of the implementation will have to be copy pasted here.

    :return: mapping of node UUIDs onto the new repository metadata.
    """
    backend = NoopRepositoryBackend()
    repository = MigrationRepository(backend=backend)

    # The original repository should be in the same parent as the DOS container, whose location can be retrieved from
    # the currently loaded storage backend.
    filepath_container = get_filepath_container(profile)
    filepath_repository = filepath_container.parent / 'repository' / 'node'
    container = Container(filepath_container)

    if not filepath_repository.exists():
        return None, None

    node_repository_dirpaths, missing_sub_repo_folder = get_node_repository_dirpaths(
        profile, filepath_repository, shard
    )

    filepaths = []
    streams = []
    mapping_metadata = {}

    # Loop over all the folders for each node that was found in the existing file repository and generate the repository
    # metadata that will have to be stored on the node. Calling `put_object_from_tree` will generate the virtual
    # hierarchy in memory, writing the files not actually to disk but opening lazy file handles, and then the call to
    # `serialize_repository` serializes the virtual hierarchy into JSON storable dictionary. This will later be stored
    # on the nodes in the database, and so it is added to the `mapping_metadata` which will be returned from this
    # function. After having constructed the virtual hierarchy, we walk over the contents and take just the files and
    # add the value (which is the `LazyOpener`) to the `streams` list as well as its relative path to `filepaths`.
    for node_uuid, node_dirpath in node_repository_dirpaths.items():
        repository.put_object_from_tree(node_dirpath)
        metadata = serialize_repository(repository)
        mapping_metadata[node_uuid] = metadata
        for root, _, filenames in repository.walk():
            for filename in filenames:
                parts = list(pathlib.Path(root / filename).parts)
                filepaths.append((node_uuid, parts))
                streams.append(functools.reduce(lambda objects, part: objects['o'].get(part), parts, metadata)['k'])

        # Reset the repository to a clean node repository, which removes the internal virtual file hierarchy
        repository.reset()

    # Free up the memory of this mapping that is no longer needed and can be big
    del node_repository_dirpaths

    hashkeys = container.add_streamed_objects_to_pack(streams, compress=False, open_streams=True)

    # Now all that remains is to go through all the generated repository metadata, stored for each node in the
    # `mapping_metadata` and replace the "values" for all the files, which are currently still the `LazyOpener`
    # instances, and replace them with the hashkey that was generated from its content by the DOS container.
    for hashkey, (node_uuid, parts) in zip(hashkeys, filepaths):
        repository_metadata = mapping_metadata[node_uuid]
        functools.reduce(lambda objects, part: objects['o'].get(part), parts, repository_metadata)['k'] = hashkey

    del filepaths
    del streams

    return mapping_metadata, missing_sub_repo_folder


def get_node_repository_dirpaths(profile, basepath, shard=None):
    """Return a mapping of node UUIDs onto the path to their current repository folder in the old repository.

    :param basepath: the absolute path of the base folder of the old file repository.
    :param shard: optional shard to define which first shard level to check. If `None`, all shard levels are checked.
    :return: dictionary of node UUID onto absolute filepath and list of node repo missing one of the two known sub
        folders, ``path`` or ``raw_input``, which is unexpected.
    :raises `~aiida.common.exceptions.StorageMigrationError`: if the repository contains node folders that contain both
        the `path` and `raw_input` subdirectories, which should never happen.
    """
    mapping = {}
    missing_sub_repo_folder = []
    contains_both = []

    if shard is not None:
        # If the shard is not present in the basepath, there is nothing to do
        if shard not in os.listdir(basepath):
            return mapping, missing_sub_repo_folder

        shards = [pathlib.Path(basepath) / shard]
    else:
        shards = basepath.iterdir()

    for shard_one in shards:
        if not REGEX_SHARD_SUB_LEVEL.match(shard_one.name):
            continue

        for shard_two in shard_one.iterdir():
            if not REGEX_SHARD_SUB_LEVEL.match(shard_two.name):
                continue

            for shard_three in shard_two.iterdir():
                if not REGEX_SHARD_FINAL_LEVEL.match(shard_three.name):
                    continue

                uuid = shard_one.name + shard_two.name + shard_three.name
                dirpath = basepath / shard_one / shard_two / shard_three
                subdirs = [path.name for path in dirpath.iterdir()]

                path = None

                if 'path' in subdirs and 'raw_input' in subdirs:
                    # If the `path` folder is empty OR it contains *only* a `.gitignore`, we simply ignore and set
                    # `raw_input` to be migrated, otherwise we add the entry to `contains_both` which will cause the
                    # migration to fail.
                    # See issue #4910 (https://github.com/aiidateam/aiida-core/issues/4910) for more information on the
                    # `.gitignore` case.
                    path_contents = os.listdir(dirpath / 'path')
                    if not path_contents or path_contents == ['.gitignore']:
                        path = dirpath / 'raw_input'
                    else:
                        contains_both.append(str(dirpath))
                elif 'path' in subdirs:
                    path = dirpath / 'path'
                elif 'raw_input' in subdirs:
                    path = dirpath / 'raw_input'
                else:
                    missing_sub_repo_folder.append(str(dirpath))

                if path is not None:
                    mapping[uuid] = path

    if contains_both and not profile.is_test_profile:
        raise exceptions.StorageMigrationError(
            f'The file repository `{basepath}` contained node repository folders that contained both the `path` as well'
            ' as the `raw_input` subfolders. This should not have happened, as the latter is used for calculation job '
            'nodes, and the former for all other nodes. The migration will be aborted and the paths of the offending '
            'node folders will be printed below. If you know which of the subpaths is incorrect, you can manually '
            'delete it and then restart the migration. Here is the list of offending node folders:\n'
            + '\n'.join(contains_both)
        )

    return mapping, missing_sub_repo_folder


def serialize_repository(repository: Repository) -> dict:
    """Serialize the metadata into a JSON-serializable format.

    .. note:: the serialization format is optimized to reduce the size in bytes.

    :return: dictionary with the content metadata.
    """
    file_object = repository._directory
    if file_object.file_type == FileType.DIRECTORY:
        if file_object.objects:
            return {'o': {key: obj.serialize() for key, obj in file_object.objects.items()}}
        return {}
    return {'k': file_object.key}


def ensure_repository_folder_created(repository_path, uuid):
    """Make sure that the repository sub folder for the node with the given UUID exists or create it.

    :param uuid: UUID of the node
    """
    dirpath = get_node_repository_sub_folder(repository_path, uuid)
    os.makedirs(dirpath, exist_ok=True)


def put_object_from_string(repository_path, uuid, name, content):
    """Write a file with the given content in the repository sub folder of the given node.

    :param uuid: UUID of the node
    :param name: name to use for the file
    :param content: the content to write to the file
    """
    ensure_repository_folder_created(repository_path, uuid)
    basepath = get_node_repository_sub_folder(repository_path, uuid)
    dirname = os.path.dirname(name)

    if dirname:
        os.makedirs(os.path.join(basepath, dirname), exist_ok=True)

    filepath = os.path.join(basepath, name)

    with open(filepath, 'w', encoding='utf-8') as handle:
        handle.write(content)


def get_node_repository_sub_folder(repository_path, uuid, subfolder='path'):
    """Return the absolute path to the sub folder `path` within the repository of the node with the given UUID.

    :param uuid: UUID of the node
    :return: absolute path to node repository folder, i.e `/some/path/repository/node/12/ab/c123134-a123/path`
    """
    uuid = str(uuid)

    repo_dirpath = os.path.join(repository_path, 'repository')
    node_dirpath = os.path.join(repo_dirpath, 'node', uuid[:2], uuid[2:4], uuid[4:], subfolder)

    return node_dirpath


def get_numpy_array_absolute_path(repository_path, uuid, name):
    """Return the absolute path of a numpy array with the given name in the repository of the node with the given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    :return: the absolute path of the numpy array file
    """
    return os.path.join(get_node_repository_sub_folder(repository_path, uuid), f'{name}.npy')


def store_numpy_array_in_repository(repository_path, uuid, name, array):
    """Store a numpy array in the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :param array: the numpy array to store
    """
    ensure_repository_folder_created(repository_path, uuid)
    filepath = get_numpy_array_absolute_path(repository_path, uuid, name)

    with open(filepath, 'wb') as handle:
        numpy.save(handle, array)


def delete_numpy_array_from_repository(repository_path, uuid, name):
    """Delete the numpy array with a given name from the repository corresponding to a node with a given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    """
    filepath = get_numpy_array_absolute_path(repository_path, uuid, name)

    try:
        os.remove(filepath)
    except OSError:
        pass


def load_numpy_array_from_repository(repository_path, uuid, name):
    """Load and return a numpy array from the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :return: the numpy array
    """
    filepath = get_numpy_array_absolute_path(repository_path, uuid, name)
    return numpy.load(filepath)


def get_repository_object(profile, hashkey):
    """Return the content of an object stored in the disk object store repository for the given hashkey."""
    filepath_container = get_filepath_container(profile)
    container = Container(filepath_container)
    return container.get_object_content(hashkey)


def recursive_datetime_to_isoformat(value):
    """Convert all datetime objects in the given value to string representations in ISO format.

    :param value: a mapping, sequence or single value optionally containing datetime objects
    """
    if isinstance(value, list):
        return [recursive_datetime_to_isoformat(_) for _ in value]

    if isinstance(value, dict):
        return dict((key, recursive_datetime_to_isoformat(val)) for key, val in value.items())

    if isinstance(value, datetime.datetime):
        return value.isoformat()

    return value


def dumps_json(dictionary):
    """Transforms all datetime object into isoformat and then returns the JSON."""
    return json.dumps(recursive_datetime_to_isoformat(dictionary))
